from copy import copy
import itertools
from typing import Any, Optional
from fandango.language.tree import DerivationTree
from fandango.constraints.constraint_visitor import ConstraintVisitor
from fandango.constraints.constraint import Constraint
from fandango.constraints.fitness import ConstraintFitness
from fandango.language.symbols.non_terminal import NonTerminal


class ConjunctionConstraint(Constraint):
    """
    Represents a conjunction constraint that can be used for fitness evaluation.
    """

    def __init__(
        self, constraints: list[Constraint], *args, lazy: bool = False, **kwargs
    ):
        """
        Initializes the conjunction constraint with the given constraints.
        :param list[Constraint] constraints: The constraints to use.
        :param args: Additional arguments.
        :param bool lazy: If True, the conjunction is lazily evaluated.
        """
        super().__init__(*args, **kwargs)
        self.constraints = constraints
        self.lazy = lazy

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on the given conjunction.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[str, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ConstraintFitness: The fitness of the tree.
        """
        tree_hash = self.get_hash(tree, scope, population, local_variables)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return copy(self.cache[tree_hash])
        if self.lazy:
            # If the conjunction is lazy, evaluate the constraints one by one and stop if one fails
            fitness_values = list()
            for constraint in self.constraints:
                fitness = constraint.fitness(tree, scope, population, local_variables)
                fitness_values.append(fitness)
                if not fitness.success:
                    break
        else:
            # If the conjunction is not lazy, evaluate all constraints at once
            fitness_values = [
                constraint.fitness(tree, scope, population, local_variables)
                for constraint in self.constraints
            ]
        # Aggregate the fitness values
        solved = sum(fitness.solved for fitness in fitness_values)
        total = sum(fitness.total for fitness in fitness_values)
        overall = all(fitness.success for fitness in fitness_values)
        failing_trees = list(
            itertools.chain.from_iterable(
                fitness.failing_trees for fitness in fitness_values
            )
        )
        if len(self.constraints) > 1:
            if overall:
                solved += 1
            total += 1
        # Create the fitness object
        fitness = ConstraintFitness(solved, total, overall, failing_trees=failing_trees)
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def format_as_spec(self) -> str:
        return "(" + " and ".join(c.format_as_spec() for c in self.constraints) + ")"

    def accept(self, visitor: "ConstraintVisitor"):
        """
        Accepts a visitor to traverse the constraint structure.
        :param ConstraintVisitor visitor: The visitor to accept.
        """
        visitor.visit_conjunction_constraint(self)
        if visitor.do_continue(self):
            for constraint in self.constraints:
                constraint.accept(visitor)



from typing import Optional, Any
from itertools import zip_longest
import random
from fandango.constraints.constraint import Constraint
from fandango.constraints.constraint_visitor import ConstraintVisitor
from fandango.constraints.fitness import ConstraintFitness, FailingTree
from fandango.constraints.failing_tree import (
    BoundsFailingTree,
    Comparison,
    ComparisonSide,
)
from fandango.errors import FandangoValueError
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.nodes.repetition import Repetition
from fandango.language.search import NonTerminalSearch
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree, index_by_reference


class RepetitionBoundsConstraint(Constraint):
    """
    Represents a constraint that checks the number of repetitions of a certain pattern in a tree.
    This is useful for ensuring that certain patterns do not occur too frequently or too infrequently.
    """

    def __init__(
        self,
        repetition_id: str,
        expr_data_min: tuple[str, list, dict],
        expr_data_max: tuple[str, list, dict],
        repetition_node: Repetition,
        *args,
        **kwargs,
    ):
        """
        Initializes the repetition bounds constraint with the given pattern and repetition bounds.
        :param NonTerminalSearch pattern: The pattern to check for repetitions.
        :param int min_reps: The minimum number of repetitions allowed.
        :param int max_reps: The maximum number of repetitions allowed.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.repetition_id = repetition_id
        self.expr_data_min = expr_data_min
        self.expr_data_max = expr_data_max
        self.search_min: Optional[NonTerminalSearch] = None
        self.search_max: Optional[NonTerminalSearch] = None
        if len(expr_data_min[1]) == 0:
            self.search_min = None
        elif len(expr_data_min[1]) == 1:
            self.search_min = expr_data_min[1][0]
        else:
            raise FandangoValueError(
                "RepetitionBoundsConstraint requires exactly one or zero searches for expr_data_max bound"
            )

        if len(expr_data_max[1]) == 0:
            self.search_max = None
        elif len(expr_data_max[1]) == 1:
            self.search_max = expr_data_max[1][0]
        else:
            raise FandangoValueError(
                "RepetitionBoundsConstraint requires exactly one or zero searches for expr_data_max bound"
            )
        self.repetition_node = repetition_node

    def _compute_rep_bound(
        self, tree_rightmost_relevant_node: "DerivationTree", expr_data
    ):
        expr, _, searches = expr_data
        local_cpy = self.local_variables.copy()

        if len(searches) == 0:
            return eval(expr, self.global_variables, local_cpy)

        nodes = []
        if len(searches) != 1:
            raise FandangoValueError(
                "Computed repetition requires exactly one or zero searches"
            )

        search_name, search = next(iter(searches.items()))
        max_path = tree_rightmost_relevant_node.get_choices_path()
        for container in search.find(tree_rightmost_relevant_node.get_root()):
            container_tree: DerivationTree = container.evaluate()
            search_in_bounds = True
            zip_var = list(zip_longest(max_path, container_tree.get_choices_path()))
            for i, (max_step, search_step) in enumerate(zip_var):
                if max_step is None:
                    break
                if search_step is None:
                    break
                if max_step.index > search_step.index:
                    break
                if max_step.index < search_step.index:
                    search_in_bounds = False
                    break
            if not search_in_bounds:
                continue
            nodes.append(container_tree)

        if len(nodes) == 0:
            raise FandangoValueError(
                f"Couldn't find search target ({search}) in prefixed DerivationTree for computed repetition"
            )

        target = nodes[-1]
        local_cpy[search_name] = target
        return eval(expr, self.global_variables, local_cpy), target

    def min(self, tree_stop_before: DerivationTree):
        return self._compute_rep_bound(tree_stop_before, self.expr_data_min)

    def max(self, tree_stop_before: DerivationTree):
        return self._compute_rep_bound(tree_stop_before, self.expr_data_max)

    def group_by_repetition_id(
        self, id_trees: list[DerivationTree]
    ) -> dict[tuple[str, int], dict[int, list[DerivationTree]]]:
        reference_trees: dict[tuple[str, int], dict[int, list[DerivationTree]]] = {}
        for id_tree in id_trees:
            iteration_ids: list[tuple[str, int, int]] = list(
                filter(lambda x: x[0] == self.repetition_id, id_tree.origin_repetitions)
            )
            for i_id in iteration_ids:
                call_id = tuple[str, int](i_id[:2])
                rep_round = i_id[2]
                # Group by id and repetition round

                if call_id not in reference_trees:
                    reference_trees[call_id] = dict()
                iter_list = reference_trees[call_id]
                if rep_round not in iter_list:
                    iter_list[rep_round] = []
                iter_list[rep_round].append(id_tree)
        return reference_trees

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on the number of repetitions of the pattern.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ConstraintFitness: The fitness of the tree.
        """
        id_trees = tree.find_by_origin(self.repetition_id)
        if len(id_trees) == 0:
            # Assume that the field containing the nr of repetitions is zero.
            # This is the case where we might have deleted all repetitions from the tree.
            return ConstraintFitness(1, 1, True)

        reference_trees = self.group_by_repetition_id(id_trees)
        failing_trees: list[FailingTree] = []
        solved = 0
        total = len(reference_trees.keys())

        for call_id in reference_trees.keys():
            iter_list = reference_trees[call_id]
            smallest_rep = min(iter_list.keys())
            highest_rep = max(iter_list.keys())
            first_iteration = iter_list[smallest_rep][0]
            last_iteration = iter_list[highest_rep][-1]

            max_bounds_search = first_iteration
            assert max_bounds_search.parent is not None
            while (
                index_by_reference(max_bounds_search.parent.children, max_bounds_search)
                == 0
            ):
                max_bounds_search = max_bounds_search.parent
                assert max_bounds_search.parent is not None

            parent = max_bounds_search.parent
            assert parent is not None
            index = index_by_reference(parent.children, max_bounds_search)
            assert (
                index is not None and index > 0
            ), "Invalid child index for bounds search"
            max_bounds_search = parent.children[index - 1]

            bound_min, min_ref_tree = self.min(max_bounds_search)
            bound_max, max_ref_tree = self.max(max_bounds_search)
            bound_len = len(iter_list)

            if bound_min <= bound_len <= bound_max:
                solved += 1
            else:
                iter_id = call_id[1]
                suggestions: list[tuple[Comparison, Any, ComparisonSide]] = []
                goal_len = random.randint(bound_min, bound_max)
                suggestions.append(
                    (
                        Comparison.EQUAL,
                        (iter_id, bound_len, goal_len),
                        ComparisonSide.RIGHT,
                    )
                )
                suggestions.append(
                    (
                        Comparison.EQUAL,
                        (iter_id, bound_len, goal_len),
                        ComparisonSide.LEFT,
                    )
                )
                assert first_iteration.parent is not None
                failing_trees.append(
                    BoundsFailingTree(
                        first_iteration.parent,
                        first_iteration,
                        last_iteration,
                        min_ref_tree,
                        max_ref_tree,
                        self,
                        suggestions=suggestions,
                    )
                )

        return ConstraintFitness(
            solved, total, solved == total, failing_trees=failing_trees
        )

    def get_first_common_node(
        self, tree_a: DerivationTree, tree_b: DerivationTree
    ) -> DerivationTree:
        common_node = tree_a.get_root(True)
        for a_path, b_path in zip(tree_a.get_choices_path(), tree_b.get_choices_path()):
            if a_path.index == b_path.index:
                common_node = common_node.children[a_path.index]
            else:
                break
        return common_node

    def fix_individual(
        self,
        grammar: Grammar,
        failing_tree: BoundsFailingTree,
        allow_repetition_full_delete: bool,
    ) -> list[tuple[DerivationTree, DerivationTree]]:
        if failing_tree.tree.read_only:
            return []
        replacements: list[tuple[DerivationTree, DerivationTree]] = list()
        for operator, value, side in failing_tree.suggestions:
            if operator == Comparison.EQUAL and side == ComparisonSide.LEFT:
                iter_id, bound_len, goal_len = value
                if goal_len > bound_len:
                    replacements.append(
                        self.insert_repetitions(
                            nr_to_insert=goal_len - bound_len,
                            rep_iteration=iter_id,
                            grammar=grammar,
                            tree=failing_tree.tree,
                            end_rep=failing_tree.ending_rep_tree,
                        )
                    )
                else:
                    if goal_len == 0 and not allow_repetition_full_delete:
                        if not allow_repetition_full_delete:
                            goal_len = 1
                    if goal_len == bound_len:
                        continue
                    delete_replace_pair = self.delete_repetitions(
                        nr_to_delete=bound_len - goal_len,
                        rep_iteration=iter_id,
                        tree=failing_tree.tree,
                    )
                    if goal_len == 0:
                        delete_replacement: DerivationTree = delete_replace_pair[1]
                        node_a = self.get_first_common_node(
                            failing_tree.tree, failing_tree.starting_rep_value
                        )
                        node_b = self.get_first_common_node(
                            failing_tree.tree, failing_tree.ending_rep_value
                        )
                        node_c = self.get_first_common_node(
                            failing_tree.starting_rep_value,
                            failing_tree.ending_rep_value,
                        )
                        # Get the node that is closest to root
                        first_node = sorted(
                            [node_a, node_b, node_c], key=lambda x: len(x.get_path())
                        )[0]
                        replacement = first_node.deepcopy(
                            copy_children=True, copy_params=False, copy_parent=False
                        )
                        replacement = replacement.replace_multiple(
                            grammar=grammar,
                            replacements=[(failing_tree.tree, delete_replacement)],
                            current_path=first_node.get_choices_path(),
                        )

                        read_only_start_idx = len(first_node.get_path()) - 1
                        current_node = replacement
                        for path_node in failing_tree.tree.get_choices_path()[
                            read_only_start_idx:
                        ]:
                            current_node = current_node.children[path_node.index]
                            current_node.read_only = True
                        current_node = replacement
                        for (
                            path_node
                        ) in failing_tree.starting_rep_value.get_choices_path()[
                            read_only_start_idx:
                        ]:
                            current_node = current_node.children[path_node.index]
                            current_node.read_only = True
                        current_node.set_all_read_only(True)
                        current_node = replacement
                        for (
                            path_node
                        ) in failing_tree.ending_rep_value.get_choices_path()[
                            read_only_start_idx:
                        ]:
                            current_node = current_node.children[path_node.index]
                            current_node.read_only = True
                        current_node.set_all_read_only(True)
                        replacements.append((first_node, replacement))
                    else:
                        replacements.append(delete_replace_pair)
                continue
        return replacements

    def insert_repetitions(
        self,
        *,
        nr_to_insert: int,
        rep_iteration: int,
        grammar: "Grammar",
        tree: DerivationTree,
        end_rep: DerivationTree,
    ) -> tuple[DerivationTree, DerivationTree]:
        assert end_rep.parent is not None, "end_rep must have a parent"
        index = index_by_reference(end_rep.parent, end_rep)
        if index is None:
            raise ValueError("end_rep not found in its parent's children")
        insertion_index = index + 1

        starting_rep = 0
        for ref in end_rep.origin_repetitions:
            if ref[0] == self.repetition_id and ref[1] == rep_iteration:
                assert ref[2] is not None, "repetition index (ref[2]) must not be None"
                starting_rep = ref[2] + 1

        old_tree_children = tree.children
        tree.set_children([])

        self.repetition_node.fuzz(
            tree,
            grammar,
            override_starting_repetition=starting_rep,
            override_current_iteration=rep_iteration,
            override_iterations_to_perform=starting_rep + nr_to_insert,
        )

        insert_children = tree.children
        tree.set_children(old_tree_children)

        copy_parent = tree.deepcopy(
            copy_children=True,
            copy_parent=False,
            copy_params=False,
        )
        copy_parent.set_children(
            copy_parent.children[:insertion_index]
            + insert_children
            + copy_parent.children[insertion_index:]
        )

        return tree, copy_parent

    def delete_repetitions(
        self, *, nr_to_delete: int, rep_iteration: int, tree: DerivationTree
    ) -> tuple[DerivationTree, DerivationTree]:
        copy_parent = tree.deepcopy(
            copy_children=True, copy_parent=False, copy_params=False
        )
        curr_rep_id = None
        reps_deleted = 0
        new_children: list[DerivationTree] = []
        for child in copy_parent.children[::-1]:
            repetition_node_id = self.repetition_id
            matching_o_nodes = list(
                filter(
                    lambda x: x[0] == repetition_node_id and x[1] == rep_iteration,
                    child.origin_repetitions,
                )
            )
            if len(matching_o_nodes) == 0:
                new_children.insert(0, child)
                continue
            matching_o_node = matching_o_nodes[0]
            rep_id = matching_o_node[2]
            if curr_rep_id != rep_id and reps_deleted >= nr_to_delete:
                # We have deleted enough repetitions iteratively add all remaining children
                new_children.insert(0, child)
                continue
            curr_rep_id = rep_id
            reps_deleted += 1
        copy_parent.set_children(new_children)
        return tree, copy_parent

    def format_as_spec(self) -> str:
        if self.search_min is None:
            print_min, _, _ = self.expr_data_min
        else:
            print_min = self.search_min.format_as_spec()
        if self.search_max is None:
            print_max, _, _ = self.expr_data_max
        else:
            print_max = self.search_max.format_as_spec()
        return f"RepetitionBounds({print_min} <= |{self.repetition_node.node.format_as_spec()}| <= {print_max})"

    def accept(self, visitor: "ConstraintVisitor"):
        """Accepts a visitor to traverse the constraint structure."""
        visitor.visit_repetition_bounds_constraint(self)



import copy
from typing import Any, Optional, TYPE_CHECKING, TypeVar, cast
from collections.abc import Iterable, Iterator
import warnings

from fandango.language.symbols import NonTerminal, Slice, Symbol, Terminal
from fandango.language.tree_value import (
    BYTES_TO_STRING_ENCODING,
    DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE,
    DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE,
    STRING_TO_BYTES_ENCODING,
    TreeValue,
    TreeValueType,
)

if TYPE_CHECKING:
    import fandango


# Recursive type for tree structure
T = TypeVar("T")
if TYPE_CHECKING:
    TreeTuple = tuple[T, list["TreeTuple[T]"]]
else:
    TreeTuple = tuple[T, list]  # beartype falls over with recursive types


class ProtocolMessage:
    """
    Holds information about a message in a protocol.
    """

    def __init__(
        self, sender: str, recipient: str | None, msg: "DerivationTree"
    ) -> None:
        self.msg = msg
        self.sender = sender
        self.recipient = recipient

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if self.recipient is None:
            return f"({self.sender} -> {self.recipient}): {str(self.msg)}"
        else:
            return f"({self.sender}): {str(self.msg)}"


class StepException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(f"StepException: {message}")


class PathStep:
    def __init__(self, index: int):
        self.index = index

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.__dict__.items())))


class SourceStep(PathStep):
    def __init__(self, index: int) -> None:
        super().__init__(index)


class ChildStep(PathStep):
    def __init__(self, index: int) -> None:
        super().__init__(index)


def index_by_reference(lst: Iterable[T], target: T) -> Optional[int]:
    for i, item in enumerate(lst):
        if item is target:  # compare reference, not data
            return i
    return None


def forward_to_tree_value_methods(
    function_names: list[str],
):
    """
    Decorator to add methods to a class, delegating to the result of `value().method(self)`.
    """

    def make_method(name):
        def method(self, *args, **kwargs):
            return getattr(self.value(), name)(*args, **kwargs)

        return method

    def decorator(cls):
        for name in function_names:
            # Check if the method is actually implemented on the class itself, not inherited
            if name in cls.__dict__:
                warnings.warn(
                    f"Method {name} already exists on {cls.__name__}, skipping",
                    Warning,
                )
            else:
                setattr(cls, name, make_method(name))
        return cls

    return decorator


@forward_to_tree_value_methods(DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE)
@forward_to_tree_value_methods(DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE)
class DerivationTree:
    """
    This class is used to represent a node in the derivation tree.
    """

    def __init__(
        self,
        symbol: Symbol,
        children: Optional[list["DerivationTree"]] = None,
        *,
        parent: Optional["DerivationTree"] = None,
        sources: Optional[list["DerivationTree"]] = None,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        read_only: Optional[bool] = False,
        origin_repetitions: Optional[list[tuple[str, int, int]]] = None,
    ) -> None:
        """
        Create a new derivation tree node.
        :param symbol: The symbol for this node (type Symbol)
        :param children: The children of this node (a list of DerivationTree)
        :param parent: (optional) The parent of this node (a DerivationTree node)
        :param sources: (optional) The sources of this node (a list of DerivationTrees used in generators to produce this node)
        :param read_only: If True, the node is read-only and cannot be modified (default: False)
        """
        if not isinstance(symbol, Symbol):
            raise TypeError(f"Expected Symbol, got {type(symbol)}")
        assert isinstance(
            symbol, (Terminal, NonTerminal, Slice)
        ), f"Received symbol of type {type(symbol)}"

        self.hash_cache: Optional[int] = None
        self._parent = parent
        self._sender = sender
        self._recipient = recipient
        self._symbol = symbol
        self._children: list[DerivationTree] = []
        self._sources: list[DerivationTree] = []  # init first
        if sources is not None:
            self.sources = sources  # use setter
        if origin_repetitions is None:
            origin_repetitions = []
        self.origin_repetitions: list[tuple[str, int, int]] = origin_repetitions
        self.read_only = read_only
        self._size = 1
        self.set_children(children or [])
        self.invalidate_hash()

    def __len__(self) -> int:
        return len(self._children)

    def count_terminals(self) -> int:
        if self.symbol.is_terminal:
            return 1
        count = 0
        for child in self._children:
            count += child.count_terminals()
        return count

    def size(self) -> int:
        return self._size

    @property
    def symbol(self) -> Terminal | NonTerminal | Slice:
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: Symbol) -> None:
        assert isinstance(
            symbol, (Terminal, NonTerminal, Slice)
        ), f"Received symbol of type {type(symbol)}"
        self._symbol = symbol
        self.invalidate_hash()

    @property
    def nonterminal(self) -> NonTerminal:
        """
        Returns the non-terminal symbol of this node.
        Raises TypeError if the symbol is not a NonTerminal.
        """
        if not self._symbol.is_non_terminal:
            raise TypeError(f"Symbol {self._symbol} is not a nonterminal")
        return cast(NonTerminal, self._symbol)

    @property
    def terminal(self) -> Terminal:
        """
        Returns the terminal symbol of this node.
        Raises TypeError if the symbol is not a Terminal.
        """
        if not self._symbol.is_terminal:
            raise TypeError(f"Symbol {self._symbol} is not a terminal")
        return cast(Terminal, self._symbol)

    @property
    def is_terminal(self) -> bool:
        """
        True is the node represents a terminal symbol.
        """
        return self.symbol.is_terminal

    @property
    def is_non_terminal(self) -> bool:
        """
        True is the node represents a nonterminal symbol.
        """
        return self.symbol.is_non_terminal

    def invalidate_hash(self) -> None:
        self.hash_cache = None
        if self._parent is not None:
            self._parent.invalidate_hash()

    @property
    def sender(self) -> Optional[str]:
        return self._sender

    @sender.setter
    def sender(self, sender: Optional[str]) -> None:
        self._sender = sender
        self.invalidate_hash()

    @property
    def recipient(self) -> Optional[str]:
        return self._recipient

    @recipient.setter
    def recipient(self, recipient: Optional[str]) -> None:
        self._recipient = recipient
        self.invalidate_hash()

    def get_path(self) -> list["DerivationTree"]:
        path: list[DerivationTree] = []
        current: Optional[DerivationTree] = self
        while current is not None:
            path.insert(0, current)
            current = current.parent
        return path

    def set_all_read_only(self, read_only: bool) -> None:
        """
        Sets self as well as all children and sources to read-only.
        This signals other classes, that this subtree should not be modified.
        """
        self.read_only = read_only
        for child in self._children:
            child.set_all_read_only(read_only)
        for child in self._sources:
            child.set_all_read_only(read_only)

    def protocol_msgs(self) -> list[ProtocolMessage]:
        """
        Returns a list of all protocol messages present in the current DerivationTree and children.
        """
        if not isinstance(self.symbol, NonTerminal):
            return []
        if self.sender is not None:
            return [ProtocolMessage(self.sender, self.recipient, self)]
        subtrees = []
        for child in self._children:
            subtrees.extend(child.protocol_msgs())
        return subtrees

    def append(
        self, hookin_path: tuple[tuple[NonTerminal, bool], ...], tree: "DerivationTree"
    ) -> None:
        """
        Appends a given DerivationTree to the current subtree at the specified hookin_path.

        :param hookin_path: A tuple of (NonTerminal, bool) pairs indicating the path to append the tree. If the bool
        is set to true, a new node is created for the NonTerminal.
        :param tree: The DerivationTree to append.
        """
        if len(hookin_path) == 0:
            self.add_child(tree)
            return
        next_nt, add_new_node = hookin_path[0]
        if add_new_node:
            self.add_child(DerivationTree(next_nt))
        elif (
            len(self.children) == 0
            or not isinstance(self.children[-1].symbol, NonTerminal)
            or self.children[-1].symbol.name() != next_nt.name()
        ):
            raise ValueError("Invalid hookin_path!")
        self.children[-1].append(hookin_path[1:], tree)

    def set_children(self, children: list["DerivationTree"]) -> None:
        self._children = children
        self._update_size(1 + sum(child.size() for child in self._children))
        for child in self._children:
            child._parent = self
        self.invalidate_hash()

    @property
    def sources(self) -> list["DerivationTree"]:
        return self._sources

    @sources.setter
    def sources(self, source: list["DerivationTree"]) -> None:
        if source is None:
            self._sources = []
        else:
            self._sources = source
        for param in self._sources:
            param._parent = self

    def add_child(self, child: "DerivationTree") -> None:
        self._children.append(child)
        self._update_size(self.size() + child.size())
        child._parent = self
        self.invalidate_hash()

    def _update_size(self, new_val: int) -> None:
        if self._parent is not None:
            self._parent._update_size(self._parent.size() + new_val - self._size)
        self._size = new_val

    def find_all_trees(self, symbol: NonTerminal) -> list["DerivationTree"]:
        trees = sum(
            [
                child.find_all_trees(symbol)
                for child in [*self._children, *self._sources]
                if child.symbol.is_non_terminal
            ],
            [],
        )
        if self.symbol == symbol:
            trees.append(self)
        return trees

    def find_direct_trees(self, symbol: NonTerminal) -> list["DerivationTree"]:
        return [
            child
            for child in [*self._children, *self._sources]
            if child.symbol == symbol
        ]

    def find_by_origin(self, node_id: str) -> list["DerivationTree"]:
        trees = sum(
            [
                child.find_by_origin(node_id)
                for child in [*self._children, *self._sources]
                if child.symbol.is_non_terminal
            ],
            [],
        )
        for o_node_id, o_iter_id, rep in self.origin_repetitions:
            if o_node_id == node_id:
                trees.append(self)
                break
        return trees

    def __getitem__(self, item: Any) -> "DerivationTree":
        if isinstance(item, list) and len(item) == 1:
            item = item[0]
        items = self._children.__getitem__(item)
        if isinstance(items, list):
            return SliceTree(items)
        else:
            return items

    def get_last_by_path(self, path: list[NonTerminal]) -> "DerivationTree":
        symbol = path[0]
        if self.symbol == symbol:
            if len(path) == 1:
                return self
            else:
                return self._get_last_by_path(path[1:])
        raise IndexError(f"No such path in tree: {path} Tree: {self}")

    def _get_last_by_path(self, path: list[NonTerminal]) -> "DerivationTree":
        symbol = path[0]
        for child in self._children[::-1]:
            if child.symbol == symbol:
                if len(path) == 1:
                    return child
                else:
                    return child._get_last_by_path(path[1:])
        raise IndexError(
            f"No such path in tree: {path} Tree: {self.get_root(stop_at_argument_begin=True)}"
        )

    def __str__(self) -> str:
        return str(self.value())

    def __int__(self) -> int:
        return int(self.value())

    def __bytes__(self) -> bytes:
        return bytes(self.value())

    def __hash__(self) -> int:
        """
        Computes a hash of the derivation tree based on its structure and symbols.
        """
        if self.hash_cache is None:
            self.hash_cache = hash(
                (
                    self.symbol,
                    self.sender,
                    self.recipient,
                    tuple(hash(child) for child in self._children),
                )
            )
        return self.hash_cache

    def __tree__(self) -> TreeTuple[Symbol]:
        return self.symbol, [child.__tree__() for child in self._children]

    @staticmethod
    def from_tree(tree: TreeTuple[str]) -> "DerivationTree":
        symbol, children = tree
        new_symbol: Symbol
        if symbol.startswith("<") and symbol.endswith(">"):
            new_symbol = NonTerminal(symbol)
        else:
            new_symbol = Terminal(symbol)
        parsed_children = [DerivationTree.from_tree(child) for child in children]
        return DerivationTree(new_symbol, parsed_children)

    def deepcopy(
        self,
        *,
        copy_children: bool = True,
        copy_params: bool = True,
        copy_parent: bool = True,
    ) -> "DerivationTree":
        return self.__deepcopy__(
            None,
            copy_children=copy_children,
            copy_params=copy_params,
            copy_parent=copy_parent,
        )

    def __deepcopy__(
        self,
        memo: Optional[dict[int, Any]],
        copy_children: bool = True,
        copy_params: bool = True,
        copy_parent: bool = True,
    ) -> "DerivationTree":
        if memo is None:
            memo = {}
        if id(self) in memo:
            res = memo[id(self)]
            assert isinstance(res, DerivationTree)
            return res

        # Create a new instance without copying the parent
        copied = DerivationTree(
            self.symbol,
            [],
            sender=self.sender,
            recipient=self.recipient,
            sources=[],
            read_only=self.read_only,
            origin_repetitions=list(self.origin_repetitions),
        )
        memo[id(self)] = copied

        # Deepcopy the children
        if copy_children:
            copied.set_children(
                [copy.deepcopy(child, memo) for child in self._children]
            )

        # Set the parent to None or update if necessary
        if copy_parent:
            copied._parent = copy.deepcopy(self.parent, memo)
        if copy_params:
            copied.sources = copy.deepcopy(self.sources, memo)

        return copied

    def should_be_serialized_to_bytes(self) -> bool:
        """
        Return true if the derivation tree should be serialized to bytes.
        """
        return self._contains_type(
            TreeValueType.TRAILING_BITS_ONLY
        ) or self._contains_type(TreeValueType.BYTES)

    def _contains_type(self, tp: TreeValueType) -> bool:
        """
        Return true if the derivation tree contains any terminal symbols of type `tp` (say, `int` or `bytes`).
        """
        if self.symbol.is_terminal and self.symbol.is_type(tp):
            return True
        return any(child._contains_type(tp) for child in self._children)

    def contains_bits(self) -> bool:
        """
        Return true iff the derivation tree contains any bits (0 or 1).
        """
        return self._contains_type(TreeValueType.TRAILING_BITS_ONLY)

    def contains_bytes(self) -> bool:
        """
        Return true iff the derivation tree contains any byte strings.
        """
        return self._contains_type(TreeValueType.BYTES)

    def to_string(self, *, encoding: str = BYTES_TO_STRING_ENCODING) -> str:
        """
        Convert the derivation tree to a string.
        """
        return self.value().to_string(bytes_to_str_encoding=encoding)

    def to_bits(self, *, encoding: str = STRING_TO_BYTES_ENCODING) -> str:
        """
        Convert the derivation tree to a sequence of bits (0s and 1s).

        """
        return self.value().to_bits(str_to_bytes_encoding=encoding)

    def to_bytes(self, encoding: str = STRING_TO_BYTES_ENCODING) -> bytes:
        """
        Convert the derivation tree to a sequence of bytes.
        String elements are encoded according to `encoding`.
        """
        return self.value().to_bytes(str_to_bytes_encoding=encoding)

    def to_int(self, encoding: str = STRING_TO_BYTES_ENCODING) -> Optional[int]:
        return self.value().to_int(str_to_bytes_encoding=encoding)

    def to_tree(self, indent: int = 0, start_indent: int = 0) -> str:
        """
        Pretty-print the derivation tree (for visualization).
        """
        s = "  " * start_indent + "Tree(" + self.symbol.format_as_spec()
        if len(self._children) == 1 and len(self._sources) == 0:
            s += ", " + self._children[0].to_tree(indent, start_indent=0)
        else:
            has_children = False
            for child in self._children:
                s += ",\n" + child.to_tree(indent + 1, start_indent=indent + 1)
                has_children = True
            if len(self._sources) > 0:
                s += ",\n" + "  " * (indent + 1) + "sources=[\n"
                for child in self._sources:
                    s += child.to_tree(indent + 2, start_indent=indent + 2) + ",\n"
                    has_children = True
                s += "  " * (indent + 1) + "]"
            if has_children:
                s += "\n" + "  " * indent
        s += ")"
        return s

    def to_repr(self, indent: int = 0, start_indent: int = 0) -> str:
        """
        Output the derivation tree in internal representation.
        """
        s = "  " * start_indent + "DerivationTree(" + self.symbol.format_as_spec()
        if len(self._children) == 1 and len(self._sources) == 0:
            s += ", [" + self._children[0].to_repr(indent, start_indent=0) + "])"
        elif len(self._children + self._sources) >= 1:
            s += ",\n" + "  " * indent + "  [\n"
            for child in self._children:
                s += child.to_repr(indent + 2, start_indent=indent + 2)
                s += ",\n"
            s += "  " * indent + "  ]\n" + "  " * indent + ")"

            if len(self._sources) > 0:
                s += ",\n" + "  " * (indent + 1) + "sources=[\n"
                for source in self._sources:
                    s += source.to_repr(indent + 2, start_indent=indent + 2)
                    s += ",\n"
                s += "  " * indent + "  ]\n" + "  " * indent + ")"
        else:
            s += ")"
        return s

    def to_grammar(
        self, include_position: bool = True, include_value: bool = True
    ) -> str:
        """
        Output the derivation tree as (specialized) grammar
        """

        def _to_grammar(
            node: "DerivationTree",
            indent: int = 0,
            start_indent: int = 0,
            bit_count: int = -1,
            byte_count: int = 0,
        ) -> tuple[str, int, int]:
            """
            Output the derivation tree as (specialized) grammar
            """
            nonlocal include_position, include_value
            assert isinstance(node.symbol, NonTerminal)

            s = "  " * start_indent + f"{node.symbol.name()} ::="
            terminal_symbols = 0

            position = f"  # Position {byte_count:#06x} ({byte_count})"
            max_bit_count = bit_count - 1

            for child in node._children:
                assert not isinstance(child.symbol, Slice)
                if child.symbol.is_non_terminal:
                    s += f" {child.symbol.format_as_spec()}"
                else:
                    terminal = cast(Terminal, child.symbol)
                    s += " " + terminal.format_as_spec()
                    terminal_symbols += 1
                    if terminal.is_type(TreeValueType.TRAILING_BITS_ONLY):
                        if bit_count <= 0:
                            bit_count = 7
                            max_bit_count = 7
                        else:
                            bit_count -= 1
                            if bit_count == 0:
                                byte_count += 1
                    else:
                        byte_count += terminal.count_bytes()
                        bit_count = -1

                # s += f" (bit_count={bit_count}, byte_count={byte_count})"

            if len(node._sources) > 0:
                # We don't know the grammar, so we report a symbolic generator
                s += (
                    " := f("
                    + ", ".join(
                        [param.symbol.format_as_spec() for param in node._sources]
                    )
                    + ")"
                )

            have_position = False
            if include_position and terminal_symbols > 0:
                have_position = True
                s += position
                if bit_count >= 0:
                    if max_bit_count != bit_count:
                        s += f", bits {max_bit_count}-{bit_count}"
                    else:
                        s += f", bit {bit_count}"

            if include_value and len(node._children) >= 2:
                s += "  # " if not have_position else "; "
                s += node.to_value()

            for child in node._children:
                if child.symbol.is_non_terminal:
                    child_str, bit_count, byte_count = _to_grammar(
                        child,
                        indent + 1,
                        start_indent=indent + 1,
                        bit_count=bit_count,
                        byte_count=byte_count,
                    )
                    s += "\n" + child_str

                for param in child._sources:
                    child_str, _, _ = _to_grammar(
                        param, indent + 2, start_indent=indent + 1
                    )
                    s += "\n  " + child_str

            return s, bit_count, byte_count

        return _to_grammar(self)[0]

    def __repr__(self) -> str:
        return self.to_repr()

    def split_end(self, copy_tree: bool = True) -> "DerivationTree":
        inst = self
        if copy_tree:
            inst = copy.deepcopy(self)
        return inst._split_end()

    def prefix(self, copy_tree: bool = True) -> "DerivationTree":
        ref_tree = self.split_end(copy_tree)
        assert ref_tree.parent is not None
        ref_tree = ref_tree.parent
        ref_tree.set_children(ref_tree.children[:-1])
        return ref_tree

    def get_root(self, stop_at_argument_begin: bool = False) -> "DerivationTree":
        root = self
        while root.parent is not None and not (
            root in root.parent.sources and stop_at_argument_begin
        ):
            root = root.parent
        return root

    def _split_end(self) -> "DerivationTree":
        if self.parent is None or self in self.parent.sources:
            if self.parent is not None:
                self._parent = None
            return self
        me_idx = index_by_reference(self.parent.children, self)
        if me_idx is None:
            # Handle error or fallback — for example:
            raise ValueError("self not found in parent's children")
        keep_children = self.parent.children[: (me_idx + 1)]
        parent = self.parent._split_end()
        parent.set_children(keep_children)
        return self

    def get_choices_path(self) -> tuple[PathStep, ...]:
        current = self
        path: list[PathStep] = []
        while current.parent is not None:
            parent = current.parent
            child_idx = index_by_reference(parent.children, current)
            if child_idx is not None:
                path.append(ChildStep(child_idx))
            else:
                source_idx = index_by_reference(parent.sources, current)
                if source_idx is None:
                    try:
                        # Fallback: If current node reference is not in parent.sources, try to get it by value.
                        source_idx = parent.sources.index(current)
                    except ValueError:
                        raise StepException(
                            f"Cannot find {current.to_repr()} in parent.children: {parent.children} or parent.sources: {parent.sources}"
                        )
                else:
                    path.append(SourceStep(source_idx))
            current = parent
        return tuple(path[::-1])

    def replace(
        self,
        grammar: "fandango.language.grammar.grammar.Grammar",  # has to be full path, otherwise beartype complains because of a circular import with grammar
        tree_to_replace: "DerivationTree",
        new_subtree: "DerivationTree",
    ) -> "DerivationTree":
        return self.replace_multiple(grammar, [(tree_to_replace, new_subtree)])

    def replace_multiple(
        self,
        grammar: "fandango.language.grammar.grammar.Grammar",  # full path to avoid circular import
        replacements: list[tuple["DerivationTree", "DerivationTree"]],
        path_to_replacement: Optional[dict[tuple, "DerivationTree"]] = None,
        current_path: Optional[tuple] = None,
    ) -> "DerivationTree":
        """
        Replace the subtree rooted at the given node with the new subtree.
        """
        if path_to_replacement is None:
            path_to_replacement = dict()
            for replacee, replacement in replacements:
                path_to_replacement[replacee.get_choices_path()] = replacement

        if current_path is None:
            current_path = self.get_choices_path()

        if (
            current_path in path_to_replacement
            and self.symbol == path_to_replacement[current_path].symbol
            and not self.read_only
        ):
            new_subtree = path_to_replacement[current_path].deepcopy(
                copy_children=True, copy_params=False, copy_parent=False
            )
            new_subtree._parent = self.parent
            new_subtree.origin_repetitions = list(self.origin_repetitions)
            new_children = []
            for i, child in enumerate(new_subtree._children):
                new_children.append(
                    child.replace_multiple(
                        grammar,
                        replacements,
                        path_to_replacement,
                        current_path + (ChildStep(i),),
                    )
                )
            new_subtree.set_children(new_children)
            grammar.populate_sources(new_subtree)
            return new_subtree

        regen_children = False
        regen_params = False
        new_children = []
        sources = []
        for i, param in enumerate(self._sources):
            new_param = param.replace_multiple(
                grammar,
                replacements,
                path_to_replacement,
                current_path + (SourceStep(i),),
            )
            sources.append(new_param)
            if new_param != param:
                regen_children = True
        for i, child in enumerate(self._children):
            new_child = child.replace_multiple(
                grammar,
                replacements,
                path_to_replacement,
                current_path + (ChildStep(i),),
            )
            new_children.append(new_child)
            if new_child != child:
                regen_params = True

        new_tree = DerivationTree(
            self.symbol,
            new_children,
            parent=self.parent,
            sender=self.sender,
            recipient=self.recipient,
            sources=sources,
            read_only=self.read_only,
            origin_repetitions=list(self.origin_repetitions),
        )

        # Update children match generator parameters, if parameters updated
        if new_tree.symbol not in grammar.generators:
            new_tree.sources = []
            return new_tree

        if regen_children:
            self_is_generator_child = False
            current = self
            current_parent = self.parent
            while current_parent is not None:
                if current in current_parent.sources:
                    break
                elif current in current_parent.children and grammar.is_use_generator(
                    current_parent
                ):
                    self_is_generator_child = True
                    break
                current = current_parent
                current_parent = current_parent.parent

            # Trees generated by generators don't contain children generated with other generators.
            if self_is_generator_child:
                new_tree.sources = []
            else:
                new_tree.set_children(grammar.derive_generator_output(new_tree))
        elif regen_params:
            new_tree.sources = grammar.derive_sources(new_tree)

        return new_tree

    def get_non_terminal_symbols(
        self, exclude_read_only: bool = True
    ) -> set[NonTerminal]:
        """
        Retrieve all non-terminal symbols present in the derivation tree.
        """
        symbols = set()
        if self.symbol.is_non_terminal and not (exclude_read_only and self.read_only):
            symbols.add(self.nonterminal)
        for child in self._children:
            symbols.update(child.get_non_terminal_symbols(exclude_read_only))
        for param in self._sources:
            symbols.update(param.get_non_terminal_symbols(exclude_read_only))
        return symbols

    def find_all_nodes(
        self, symbol: NonTerminal, exclude_read_only: bool = True
    ) -> list["DerivationTree"]:
        """
        Find all nodes in the derivation tree with the given non-terminal symbol.
        """
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        nodes = []
        if self.symbol.is_non_terminal:
            nt = cast(NonTerminal, self.symbol)
            if nt == symbol and not (exclude_read_only and self.read_only):
                nodes.append(self)
            for child in self._children:
                nodes.extend(child.find_all_nodes(symbol, exclude_read_only))
            for param in self._sources:
                nodes.extend(param.find_all_nodes(symbol, exclude_read_only))
        return nodes

    @property
    def children(self) -> list["DerivationTree"]:
        """
        Return the children of the current node.
        """
        return self._children

    @property
    def parent(self) -> Optional["DerivationTree"]:
        """
        Return the parent node of the current node.
        """
        return self._parent

    def children_values(self) -> list[TreeValue]:
        """
        Return values of all direct children
        """
        return [node.value() for node in self.children]

    def flatten(self) -> list["DerivationTree"]:
        """
        Flatten the derivation tree into a list of DerivationTrees.
        """
        flat = [self]
        for child in self._children:
            flat.extend(child.flatten())
        return flat

    def descendants(self) -> list["DerivationTree"]:
        """
        Return all descendants of the current node
        """
        return self.flatten()[1:]

    def descendant_values(self) -> list[TreeValue]:
        """
        Return all descendants of the current node
        """
        values = [node.value() for node in self.descendants()]
        return values

    def get_index(self, target: "DerivationTree") -> int:
        """
        Get the index of the target node in the tree.
        """
        flat = self.flatten()
        try:
            return flat.index(target)
        except ValueError:
            return -1

    def value(self) -> TreeValue:
        if self.symbol.is_terminal:
            return self.symbol.value()

        aggregate = TreeValue.empty()
        for child in self._children:
            aggregate = aggregate.append(child.value())
        return aggregate

    def to_value(self) -> str:
        value = self.value()
        if value.is_type(TreeValueType.EMPTY):
            return ""
        elif value.is_type(TreeValueType.TRAILING_BITS_ONLY):
            return "0b" + value.to_bits()
        elif value.is_type(TreeValueType.STRING):
            return str(value)
        elif value.is_type(TreeValueType.BYTES):
            return str(bytes(value))
        else:
            raise ValueError(f"Invalid value type: {value.type_}")

    ## Comparison operations
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DerivationTree):
            return hash(self) == hash(other)
        else:
            return self.value() == other

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __iter__(self) -> Iterator["DerivationTree"]:
        return iter(self._children)


class SliceTree(DerivationTree):
    def __init__(self, children: list[DerivationTree], read_only: bool = False) -> None:
        super().__init__(Slice(), children, read_only=read_only)



import ast
from collections.abc import Sequence
from io import UnsupportedOperation
from typing import Any, Optional

from fandango.constraints.soft import SoftValue
from fandango.constraints.repetition_bounds import RepetitionBoundsConstraint
from fandango.errors import FandangoValueError
from fandango.constraints.constraint import Constraint
from fandango.constraints.comparison import ComparisonConstraint
from fandango.constraints.conjunction import ConjunctionConstraint
from fandango.constraints.disjunct import DisjunctionConstraint
from fandango.constraints.exists import ExistsConstraint
from fandango.constraints.expression import ExpressionConstraint
from fandango.constraints.forall import ForallConstraint
from fandango.constraints.failing_tree import Comparison
from fandango.language import NonTerminalSearch, NodeType
from fandango.language.grammar import FuzzingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.grammar_settings import GrammarSetting
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.char_set import CharSet
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Option, Plus, Repetition, Star
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.parser.FandangoParser import FandangoParser
from fandango.language.parser.FandangoParserVisitor import FandangoParserVisitor
from fandango.language.search import (
    AttributeSearch,
    DescendantAttributeSearch,
    ItemSearch,
    LengthSearch,
    RuleSearch,
    SelectiveSearch,
    StarSearch,
    PopulationSearch,
)
from fandango.language.symbols import NonTerminal, Terminal
from fandango.logger import LOGGER


class FandangoSplitter(FandangoParserVisitor):
    def __init__(self):
        self.productions = []
        self.constraints = []
        self.grammar_settings: list[FandangoParser.Grammar_setting_contentContext] = []
        self.python_code = []

    def visitFandango(self, ctx: FandangoParser.FandangoContext):
        self.productions = []
        self.constraints = []
        self.grammar_settings = []
        self.python_code = []
        self.visitChildren(ctx)

    def visitProduction(self, ctx: FandangoParser.ProductionContext):
        self.productions.append(ctx)

    def visitConstraint(self, ctx: FandangoParser.ConstraintContext):
        self.constraints.append(ctx)

    def visitGrammar_setting_content(
        self, ctx: FandangoParser.Grammar_setting_contentContext
    ):
        self.grammar_settings.append(ctx)

    def visitPython(self, ctx: FandangoParser.PythonContext):
        self.python_code.append(ctx)


class GrammarProcessor(FandangoParserVisitor):
    def __init__(
        self,
        grammar_settings: list[FandangoParser.Grammar_setting_contentContext],
        local_variables: Optional[dict[str, Any]] = None,
        global_variables: Optional[dict[str, Any]] = None,
        id_prefix: Optional[str] = None,
        max_repetitions: int = 5,
    ):
        self.local_variables = local_variables
        self.global_variables = global_variables
        self.id_prefix = id_prefix
        self.searches = SearchProcessor(Grammar.dummy())
        self.repetition_constraints: list[RepetitionBoundsConstraint] = list()
        self.seenParties = set[str]()
        self.additionalRules = dict[NonTerminal, Node]()
        self.max_repetitions = max_repetitions
        self._grammar_settings: Sequence[HasSettings] = [
            self.visitGrammar_setting_content(ctx) for ctx in grammar_settings
        ]

        if self.id_prefix is None:
            self.id_prefix = ""

        self.seenAlternatives = 0
        self.seenConcatenations = 0
        self.seenStars = 0
        self.seenRepetitions = 0
        self.seenOptions = 0
        self.seenPluses = 0

    def get_grammar(
        self,
        productions: list[FandangoParser.ProductionContext],
        prime: bool = True,
    ) -> Grammar:
        grammar = Grammar(
            grammar_settings=self._grammar_settings,
            local_variables=self.local_variables,
            global_variables=self.global_variables,
        )
        for production in productions:
            symbol = NonTerminal(production.nonterminal().getText())
            grammar[symbol] = self.visitAlternative(production.alternative())
            if grammar.has_generator(symbol):
                grammar.remove_generator(symbol)
            if production.expression():
                # Handle generator expressions
                expr, _, searches_map = self.searches.visit(production.expression())
                grammar.set_generator(symbol, ast.unparse(expr), searches_map)

                if not production.EXPR_ASSIGN():
                    LOGGER.warning(
                        f"{symbol.format_as_spec()}: Using '=' and '::' for generators is deprecated. Use ':=' instead."
                    )
            if production.SEMI_COLON():
                LOGGER.info(
                    f"{symbol.format_as_spec()}: A final ';' is not required in grammar rules."
                )

        grammar.rules.update(self.additionalRules)
        grammar.fuzzing_mode = (
            FuzzingMode.COMPLETE if len(self.seenParties) == 0 else FuzzingMode.IO
        )
        grammar.update_parser()
        if prime:
            grammar.prime()
        return grammar

    def visitGrammar_setting_content(
        self, ctx: FandangoParser.Grammar_setting_contentContext
    ) -> GrammarSetting:
        selector = ctx.grammar_selector().getText()
        unparsed_rules: list[FandangoParser.Grammar_ruleContext] = ctx.grammar_rule()
        rules = {
            rule.grammar_setting_key().getText(): rule.grammar_setting_value().getText()
            for rule in unparsed_rules
        }
        return GrammarSetting(selector, rules)

    def visitAlternative(self, ctx: FandangoParser.AlternativeContext):
        nodes = [self.visitConcatenation(child) for child in ctx.concatenation()]
        if len(nodes) == 1:
            return nodes[0]
        self.seenAlternatives += 1
        nid = self.seenAlternatives
        return Alternative(
            nodes,
            self._grammar_settings,
            f"{NodeType.ALTERNATIVE}:{nid}_{self.id_prefix}",
        )

    def visitConcatenation(self, ctx: FandangoParser.ConcatenationContext):
        nodes = [self.visitOperator(child) for child in ctx.operator()]
        if len(nodes) == 1:
            return nodes[0]
        self.seenConcatenations += 1
        nid = self.seenConcatenations
        return Concatenation(
            nodes,
            self._grammar_settings,
            f"{NodeType.CONCATENATION}:{nid}_{self.id_prefix}",
        )

    def visitKleene(self, ctx: FandangoParser.KleeneContext):
        self.seenStars += 1
        nid = self.seenStars
        return Star(
            self.visit(ctx.symbol()),
            self._grammar_settings,
            f"{NodeType.STAR}:{nid}_{self.id_prefix}",
        )

    def visitPlus(self, ctx: FandangoParser.PlusContext):
        self.seenPluses += 1
        nid = self.seenPluses
        return Plus(
            self.visit(ctx.symbol()),
            self._grammar_settings,
            f"{NodeType.PLUS}:{nid}_{self.id_prefix}",
        )

    def visitOption(self, ctx: FandangoParser.OptionContext):
        self.seenOptions += 1
        nid = self.seenOptions
        return Option(
            self.visit(ctx.symbol()),
            self._grammar_settings,
            f"{NodeType.OPTION}:{nid}_{self.id_prefix}",
        )

    def visitRepeat(self, ctx: FandangoParser.RepeatContext):
        node = self.visitSymbol(ctx.symbol())
        self.seenRepetitions += 1
        nid = f"{NodeType.REPETITION}:{self.seenRepetitions}_{self.id_prefix}"
        min_ = None
        max_ = None
        if ctx.COMMA():
            bounds_index = 0
            started = False
            for child in ctx.getChildren():
                if child.getText() == "{":
                    started = True
                elif child.getText() == "}":
                    break
                elif started:
                    if child.getText() == ",":
                        bounds_index += 1
                    else:
                        search_bound = self.searches.visit(child)
                        if bounds_index == 0:
                            min_ = (
                                ast.unparse(search_bound[0]),
                                search_bound[1],
                                search_bound[2],
                            )
                        else:
                            max_ = (
                                ast.unparse(search_bound[0]),
                                search_bound[1],
                                search_bound[2],
                            )

            max_arg = None
            min_arg = 0
            require_constraint = False
            bounds_constraint = None
            if max_ is not None:
                if max_[0].isdigit():
                    max_arg = int(max_[0])
                else:
                    require_constraint = True
            else:
                max_ = (f"{self.max_repetitions}", [], {})

            if min_ is not None:
                if min_[0].isdigit():
                    min_arg = int(min_[0])
                else:
                    require_constraint = True
            else:
                min_ = (f"{min_arg}", [], {})
            if require_constraint:
                if min_arg == 0:
                    min_arg = 1
                if max_arg is not None and max_arg < min_arg:
                    max_arg = min_arg
            rep_node = Repetition(
                node, self._grammar_settings, nid, min_=min_arg, max_=max_arg
            )
            if require_constraint:
                bounds_constraint = RepetitionBoundsConstraint(
                    nid,
                    expr_data_min=min_,
                    expr_data_max=max_,
                    repetition_node=rep_node,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                )
                self.repetition_constraints.append(bounds_constraint)
            if bounds_constraint is not None:
                bounds_constraint.repetition_node = rep_node
            return rep_node
        reps_visit = self.searches.visit(ctx.expression(0))
        reps: tuple[str, list, dict] = (ast.unparse(reps_visit[0]), *reps_visit[1:])
        if reps[0].isdigit():
            return Repetition(
                node, self._grammar_settings, nid, int(reps[0]), int(reps[0])
            )
        else:
            rep_node = Repetition(node, self._grammar_settings, nid, min_=1)
            bounds_constraint = RepetitionBoundsConstraint(
                nid,
                expr_data_min=reps,
                expr_data_max=reps,
                repetition_node=rep_node,
                local_variables=self.local_variables,
                global_variables=self.global_variables,
            )
            self.repetition_constraints.append(bounds_constraint)
            rep_node.bounds_constraint = bounds_constraint

            return rep_node

    def visitSymbol(self, ctx: FandangoParser.SymbolContext):
        if ctx.nonterminal_right():
            return self.visitNonterminal_right(ctx.nonterminal_right())
        elif ctx.string():
            return TerminalNode(
                Terminal.from_symbol(ctx.string().getText()), self._grammar_settings
            )
        elif ctx.NUMBER():
            number = ctx.NUMBER().getText()
            if number not in ["0", "1"]:
                raise UnsupportedOperation(f"Unsupported bit spec: {number}")
            return TerminalNode(Terminal.from_number(number), self._grammar_settings)
        elif ctx.char_set():
            text = ctx.char_set().getText()
            LOGGER.warning(
                f"{text}: Charset specs are deprecated. Use regular expressions (r'...') instead."
            )
            return CharSet(text, self._grammar_settings)
        elif ctx.alternative():
            return self.visitAlternative(ctx.alternative())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitNonterminal_right(self, ctx: FandangoParser.Nonterminal_rightContext):
        if ctx.identifier(1) is None:
            return NonTerminalNode(
                NonTerminal("<" + ctx.identifier(0).getText() + ">"),
                self._grammar_settings,
            )
        elif ctx.identifier(2) is None:
            self.seenParties.add(ctx.identifier(0).getText())
            return NonTerminalNode(
                NonTerminal("<" + ctx.identifier(1).getText() + ">"),
                self._grammar_settings,
                ctx.identifier(0).getText(),
                None,
            )
        else:
            self.seenParties.add(ctx.identifier(0).getText())
            self.seenParties.add(ctx.identifier(1).getText())
            return NonTerminalNode(
                NonTerminal("<" + ctx.identifier(2).getText() + ">"),
                self._grammar_settings,
                ctx.identifier(0).getText(),
                ctx.identifier(1).getText(),
            )


class ConstraintProcessor(FandangoParserVisitor):
    def __init__(
        self,
        grammar: Grammar,
        local_variables: Optional[dict[str, Any]] = None,
        global_variables: Optional[dict[str, Any]] = None,
        lazy: bool = False,
    ):
        self.searches = SearchProcessor(grammar)
        self.lazy = lazy
        self.local_variables = local_variables
        self.global_variables = global_variables

    def get_constraints(
        self, constraints: list[FandangoParser.ConstraintContext]
    ) -> list[Constraint | SoftValue]:
        return [self.visitConstraint(constraint) for constraint in constraints]
        # if len(constraints) == 1:
        #     return constraints[0]
        # else:
        #     return ConjunctionConstraint(
        #         constraints,
        #         local_variables=self.local_variables,
        #         global_variables=self.global_variables,
        #         lazy=self.lazy,
        #    )

    def visitConstraint(self, ctx: FandangoParser.ConstraintContext):
        LOGGER.debug(f"Visiting constraint: {ctx.getText()}")
        if (not ctx.WHERE()) and (not ctx.MINIMIZING()) and (not ctx.MAXIMIZING()):
            LOGGER.warning(
                f"{ctx.getText()}: Constraints should be prefixed with 'where'."
            )
        if (ctx.WHERE() or ctx.MINIMIZING() or ctx.MAXIMIZING()) and ctx.SEMI_COLON():
            LOGGER.info(
                f"{ctx.getText()}: A final ';' is not required with 'where', 'minimizing', 'maximizing'."
            )
        if ctx.implies():
            constraint = self.visitImplies(ctx.implies())
            return constraint
        elif ctx.MINIMIZING() or ctx.MAXIMIZING():
            expression_constraint = self.visitExpr(ctx.expr())
            optimization_goal = "min" if ctx.MINIMIZING() else "max"
            return SoftValue(
                optimization_goal,
                expression_constraint.expression,
                searches=expression_constraint.searches,
                local_variables=expression_constraint.local_variables,
                global_variables=expression_constraint.global_variables,
            )
        else:
            raise FandangoValueError(f"Unknown constraint: {ctx.getText()}")

    def visitImplies(self, ctx: FandangoParser.ImpliesContext):
        if ctx.ARROW():
            operands = ctx.getText().split("->")
            LOGGER.warning(
                f"{ctx.getText()}: Implication is deprecated. Use `not({operands[0]}) or {operands[1]}` instead."
            )
        return self.visitQuantifier(ctx.quantifier())

    def visitQuantifier(self, ctx: FandangoParser.QuantifierContext):
        if ctx.formula_disjunction():
            return self.visitFormula_disjunction(ctx.formula_disjunction())
        elif ctx.EXISTS() or ctx.FORALL():
            constraint = self.visitQuantifier(ctx.quantifier())
            bound = NonTerminal(ctx.nonterminal().getText())
            search = self.searches.visitDot_selection(ctx.dot_selection())[1][0]
            if ctx.EXISTS():
                return ExistsConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            elif ctx.FORALL():
                return ForallConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            else:
                raise FandangoValueError(f"Unknown quantifier: {ctx.getText()}")
        elif ctx.ANY() or ctx.ALL():
            constraint = self.visitQuantifier_in_line(ctx.quantifier_in_line())
            if ctx.nonterminal():
                bound = NonTerminal(ctx.nonterminal().getText())
            else:
                bound = ctx.identifier().getText()
            search = self.searches.visitStar_selection(ctx.star_selection())[1][0]
            if ctx.ANY():
                return ExistsConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            elif ctx.ALL():
                return ForallConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            else:
                raise FandangoValueError(f"Unknown quantifier: {ctx.getText()}")
        else:
            raise FandangoValueError(f"Unknown quantifier: {ctx.getText()}")

    def visitQuantifier_in_line(self, ctx: FandangoParser.Quantifier_in_lineContext):
        if ctx.formula_disjunction():
            return self.visitFormula_disjunction(ctx.formula_disjunction())
        elif ctx.ANY() or ctx.ALL():
            constraint = self.visitQuantifier_in_line(ctx.quantifier_in_line())
            if ctx.nonterminal():
                bound = NonTerminal(ctx.nonterminal().getText())
            else:
                bound = ctx.identifier().getText()
            search = self.searches.visitStar_selection(ctx.star_selection())[1][0]
            if ctx.ANY():
                return ExistsConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            elif ctx.ALL():
                return ForallConstraint(
                    constraint,
                    bound,
                    search,
                    local_variables=self.local_variables,
                    global_variables=self.global_variables,
                    lazy=self.lazy,
                )
            else:
                raise FandangoValueError(f"Unknown quantifier: {ctx.getText()}")
        else:
            raise FandangoValueError(f"Unknown quantifier: {ctx.getText()}")

    def visitFormula_disjunction(self, ctx: FandangoParser.Formula_disjunctionContext):
        constraints = [
            self.visitFormula_conjunction(constraint)
            for constraint in ctx.formula_conjunction()
        ]
        if len(constraints) == 1:
            return constraints[0]
        else:
            return DisjunctionConstraint(
                constraints,
                local_variables=self.local_variables,
                global_variables=self.global_variables,
                lazy=self.lazy,
            )

    def visitFormula_conjunction(self, ctx: FandangoParser.Formula_conjunctionContext):
        constraints = [
            self.visitFormula_atom(constraint) for constraint in ctx.formula_atom()
        ]
        if len(constraints) == 1:
            return constraints[0]
        else:
            return ConjunctionConstraint(
                constraints,
                local_variables=self.local_variables,
                global_variables=self.global_variables,
                lazy=self.lazy,
            )

    def visitFormula_atom(self, ctx: FandangoParser.Formula_atomContext):
        if ctx.implies():
            return self.visitImplies(ctx.implies())
        elif ctx.expr():
            return self.visitExpr(ctx.expr())
        elif ctx.formula_comparison():
            return self.visitFormula_comparison(ctx.formula_comparison())
        else:
            raise FandangoValueError(f"Unknown formula atom: {ctx.getText()}")

    def visitExpr(self, ctx: FandangoParser.ExprContext):
        if ctx.selector_length():
            expr, _, search_map = self.searches.visitSelector_length(
                ctx.selector_length()
            )
        elif ctx.IF():
            body, _, body_search_map = self.searches.visitInversion(ctx.inversion(0))
            test, _, test_search_map = self.searches.visitInversion(ctx.inversion(1))
            orelse, _, orelse_search_map = self.searches.visitInversion(
                ctx.inversion(2)
            )
            search_map = {**body_search_map, **test_search_map, **orelse_search_map}
            expr = ast.IfExp(test=test, body=body, orelse=orelse)
        elif ctx.inversion():
            expr, _, search_map = self.searches.visitInversion(ctx.inversion(0))
        else:
            raise FandangoValueError(f"Unknown expression: {ctx.getText()}")
        return ExpressionConstraint(
            ast.unparse(expr),
            searches=search_map,
            local_variables=self.local_variables,
            global_variables=self.global_variables,
        )

    def visitFormula_comparison(self, ctx: FandangoParser.Formula_comparisonContext):
        if ctx.LESS_THAN():
            op = Comparison.LESS
        elif ctx.GREATER_THAN():
            op = Comparison.GREATER
        elif ctx.EQUALS():
            op = Comparison.EQUAL
        elif ctx.GT_EQ():
            op = Comparison.GREATER_EQUAL
        elif ctx.LT_EQ():
            op = Comparison.LESS_EQUAL
        elif ctx.NOT_EQ_1() or ctx.NOT_EQ_2():
            op = Comparison.NOT_EQUAL
        else:
            raise UnsupportedOperation(f"Unknown operator in {ctx.getText()}")
        left, _, left_map = self.searches.visit(ctx.expr(0))
        right, _, right_map = self.searches.visit(ctx.expr(1))
        return ComparisonConstraint(
            op,
            ast.unparse(left),
            ast.unparse(right),
            searches={**left_map, **right_map},
            local_variables=self.local_variables,
            global_variables=self.global_variables,
        )


class SearchProcessor(FandangoParserVisitor):
    def __init__(self, grammar: Grammar):
        self.identifier_id = 0
        self.grammar = grammar

    def defaultResult(
        self,
    ) -> tuple[
        ast.AST | list[ast.AST], list[AttributeSearch], dict[str, AttributeSearch]
    ]:
        return [], [], {}

    # noinspection PyPep8Naming
    def aggregateResult(self, aggregate, nextResult):
        tree_aggregate, searches_aggregate, search_map_aggregate = aggregate
        tree_next, searches_next, search_map_next = nextResult
        searches = searches_aggregate + searches_next
        search_map = {**search_map_aggregate, **search_map_next}
        if isinstance(tree_aggregate, list):
            if isinstance(tree_next, list):
                tree = tree_aggregate + tree_next
            else:
                tree = tree_aggregate + [tree_next]
        elif isinstance(tree_next, list):
            tree = [tree_aggregate] + tree_next
        else:
            tree = [tree_aggregate, tree_next]
        return tree, searches, search_map

    def get_new_identifier(self):
        identifier = f"___fandango_{id(self)}_{self.identifier_id}___"
        self.identifier_id += 1
        return identifier

    def visitRs_pairs(self, ctx: FandangoParser.Rs_pairsContext):
        symbols = list()
        slices = list()
        for child in ctx.rs_pair():
            sym, items = self.visitRs_pair(child)
            symbols.append(sym)
            slices.append(items)
        return symbols, slices

    def visitRs_pair(self, ctx: FandangoParser.Rs_pairContext):
        symbol = (
            NonTerminal(ctx.nonterminal().getText()),
            False if ctx.STAR() else True,
        )
        if ctx.rs_slice():
            return symbol, self.visitRs_slice(ctx.rs_slice())
        return symbol, None

    def visitRs_slice(self, ctx: FandangoParser.Rs_sliceContext):
        if ctx.COLON():
            return slice(
                int(ctx.NUMBER(0).getText()) if ctx.NUMBER(0) else None,
                int(ctx.NUMBER(1).getText()) if ctx.NUMBER(1) else None,
                int(ctx.NUMBER(2).getText()) if ctx.NUMBER(2) else None,
            )
        else:
            return int(ctx.NUMBER(0).getText())

    def visitRs_slices(self, ctx: FandangoParser.Rs_slicesContext):
        return [self.visitRs_slice(child) for child in ctx.rs_slice()]

    def visitBase_selection(self, ctx: FandangoParser.Base_selectionContext):
        if ctx.nonterminal():
            return RuleSearch(NonTerminal(ctx.nonterminal().getText()))
        elif ctx.dot_selection():
            return self.get_attribute_searches(ctx.dot_selection())
        else:
            raise FandangoValueError(f"Unknown base selection: {ctx.getText()}")

    def transform_selection(self, ctx: FandangoParser.SelectionContext):
        base = self.visitBase_selection(ctx.base_selection())
        if ctx.rs_pairs():
            symbols, slices = self.visitRs_pairs(ctx.rs_pairs())
            return SelectiveSearch(base, symbols, slices)
        elif ctx.rs_slices():
            return ItemSearch(base, self.visitRs_slices(ctx.rs_slices()))
        return base

    def get_attribute_searches(self, ctx: FandangoParser.Dot_selectionContext):
        search = self.transform_selection(ctx.selection())

        if ctx.DOT():
            return AttributeSearch(
                self.get_attribute_searches(ctx.dot_selection()), search
            )
        elif ctx.DOTDOT():
            return DescendantAttributeSearch(
                self.get_attribute_searches(ctx.dot_selection()), search
            )
        else:
            return search

    def visitSelector_length(self, ctx: FandangoParser.Selector_lengthContext):
        if ctx.OR_OP():
            tree, searches, search_map = self.visitDot_selection(ctx.dot_selection())
            search_map[tree.id] = LengthSearch(searches[0])
        elif ctx.LEN():
            tree, searches, search_map = self.visitStar_selection(ctx.star_selection())
            search_map[tree.id] = LengthSearch(searches[0])
        else:
            tree, searches, search_map = self.visitStar_selection_or_dot_selection(
                ctx.star_selection_or_dot_selection()
            )
        return tree, searches, search_map

    def visitStar_selection_or_dot_selection(
        self, ctx: FandangoParser.Star_selection_or_dot_selectionContext
    ):
        if ctx.star_selection():
            return self.visitStar_selection(ctx.star_selection())
        elif ctx.dot_selection():
            return self.visitDot_selection(ctx.dot_selection())
        else:
            raise FandangoValueError(f"Unknown selection: {ctx.getText()}")

    def visitStar_selection(self, ctx: FandangoParser.Star_selectionContext):
        identifier = self.get_new_identifier()
        base = self.get_attribute_searches(ctx.dot_selection())
        search: NonTerminalSearch
        if ctx.STAR():
            search = StarSearch(base)
        elif ctx.POWER():
            search = PopulationSearch(base)
        else:
            raise FandangoValueError(f"Unknown star selection: {ctx.getText()}")
        return ast.Name(id=identifier), [search], {identifier: search}

    def visitDot_selection(self, ctx: FandangoParser.Dot_selectionContext):
        identifier = self.get_new_identifier()
        search = self.get_attribute_searches(ctx)
        return ast.Name(id=identifier), [search], {identifier: search}

    def visitExpression(self, ctx: FandangoParser.ExpressionContext):
        if ctx.IF():
            then_ast, then_searches, then_map = self.visitDisjunction(
                ctx.disjunction(0)
            )
            test_ast, test_searches, test_map = self.visitDisjunction(
                ctx.disjunction(1)
            )
            else_ast, else_searches, else_map = self.visitExpression(ctx.expression())
            return (
                ast.IfExp(
                    test=test_ast,
                    body=then_ast,
                    orelse=else_ast,
                ),
                then_searches + test_searches + else_searches,
                {**then_map, **test_map, **else_map},
            )
        elif ctx.disjunction():
            return self.visitDisjunction(ctx.disjunction(0))
        else:
            return self.visitLambdef(ctx.lambdef())

    def visitDisjunction(self, ctx: FandangoParser.DisjunctionContext):
        if ctx.OR():
            trees, searches, search_map = self.visitChildren(ctx)
            return (
                ast.BoolOp(
                    op=ast.Or(),
                    values=trees,
                ),
                searches,
                search_map,
            )
        return self.visitConjunction(ctx.conjunction(0))

    def visitConjunction(self, ctx: FandangoParser.ConjunctionContext):
        if ctx.AND():
            trees, searches, search_map = self.visitChildren(ctx)
            return (
                ast.BoolOp(
                    op=ast.And(),
                    values=trees,
                ),
                searches,
                search_map,
            )
        return self.visitInversion(ctx.inversion(0))

    def _visit_unary_op(self, ctx, op):
        trees, searches, search_map = self.visitChildren(ctx)
        return (
            ast.UnaryOp(
                op=op,
                operand=trees[0],
            ),
            searches,
            search_map,
        )

    def visitInversion(self, ctx: FandangoParser.InversionContext):
        if ctx.NOT():
            return self._visit_unary_op(ctx, ast.Not())
        return self.visitComparison(ctx.comparison())

    def visitComparison(self, ctx: FandangoParser.ComparisonContext):
        if ctx.compare_op_bitwise_or_pair():
            left, searches, search_map = self.visitBitwise_or(ctx.bitwise_or())
            operators = list()
            trees = list()
            for comparison in ctx.compare_op_bitwise_or_pair():
                operator, result = self.visitCompare_op_bitwise_or_pair(comparison)
                r_tree, r_searches, r_search_map = result
                operators.append(operator)
                trees.append(r_tree)
                searches.extend(r_searches)
                search_map.update(r_search_map)
            return (
                ast.Compare(
                    left=left,
                    ops=operators,
                    comparators=trees,
                ),
                searches,
                search_map,
            )
        return self.visitBitwise_or(ctx.bitwise_or())

    def visitCompare_op_bitwise_or_pair(
        self, ctx: FandangoParser.Compare_op_bitwise_or_pairContext
    ):
        return self.visit(ctx.children[0])

    def visitEq_bitwise_or(self, ctx: FandangoParser.Eq_bitwise_orContext):
        return ast.Eq(), self.visitBitwise_or(ctx.bitwise_or())

    def visitNoteq_bitwise_or(self, ctx: FandangoParser.Noteq_bitwise_orContext):
        return ast.NotEq(), self.visitBitwise_or(ctx.bitwise_or())

    def visitLte_bitwise_or(self, ctx: FandangoParser.Lte_bitwise_orContext):
        return ast.LtE(), self.visitBitwise_or(ctx.bitwise_or())

    def visitGte_bitwise_or(self, ctx: FandangoParser.Gte_bitwise_orContext):
        return ast.GtE(), self.visitBitwise_or(ctx.bitwise_or())

    def visitLt_bitwise_or(self, ctx: FandangoParser.Lt_bitwise_orContext):
        return ast.Lt(), self.visitBitwise_or(ctx.bitwise_or())

    def visitGt_bitwise_or(self, ctx: FandangoParser.Gt_bitwise_orContext):
        return ast.Gt(), self.visitBitwise_or(ctx.bitwise_or())

    def visitNotin_bitwise_or(self, ctx: FandangoParser.Notin_bitwise_orContext):
        return ast.NotIn(), self.visitBitwise_or(ctx.bitwise_or())

    def visitIn_bitwise_or(self, ctx: FandangoParser.In_bitwise_orContext):
        return ast.In(), self.visitBitwise_or(ctx.bitwise_or())

    def visitIs_bitwise_or(self, ctx: FandangoParser.Is_bitwise_orContext):
        return ast.Is(), self.visitBitwise_or(ctx.bitwise_or())

    def visitIsnot_bitwise_or(self, ctx: FandangoParser.Isnot_bitwise_orContext):
        return ast.IsNot(), self.visitBitwise_or(ctx.bitwise_or())

    def _visit_bin_op(self, ctx, op):
        trees, searches, search_map = self.visitChildren(ctx)
        return (
            ast.BinOp(
                left=trees[0],
                op=op,
                right=trees[1],
            ),
            searches,
            search_map,
        )

    def visitBitwise_or(self, ctx: FandangoParser.Bitwise_orContext):
        if ctx.bitwise_or():
            return self._visit_bin_op(ctx, ast.BitOr())
        return self.visitBitwise_xor(ctx.bitwise_xor())

    def visitBitwise_xor(self, ctx: FandangoParser.Bitwise_xorContext):
        if ctx.bitwise_xor():
            return self._visit_bin_op(ctx, ast.BitXor())
        return self.visitBitwise_and(ctx.bitwise_and())

    def visitBitwise_and(self, ctx: FandangoParser.Bitwise_andContext):
        if ctx.bitwise_and():
            return self._visit_bin_op(ctx, ast.BitAnd())
        return self.visitShift_expr(ctx.shift_expr())

    def visitShift_expr(self, ctx: FandangoParser.Shift_exprContext):
        if ctx.LEFT_SHIFT():
            return self._visit_bin_op(ctx, ast.LShift())
        elif ctx.RIGHT_SHIFT():
            return self._visit_bin_op(ctx, ast.RShift())
        return self.visitSum(ctx.sum_())

    def visitSum(self, ctx: FandangoParser.SumContext):
        if ctx.ADD():
            return self._visit_bin_op(ctx, ast.Add())
        elif ctx.MINUS():
            return self._visit_bin_op(ctx, ast.Sub())
        return self.visitTerm(ctx.term())

    def visitTerm(self, ctx: FandangoParser.TermContext):
        if ctx.STAR():
            return self._visit_bin_op(ctx, ast.Mult())
        elif ctx.DIV():
            return self._visit_bin_op(ctx, ast.Div())
        elif ctx.IDIV():
            return self._visit_bin_op(ctx, ast.FloorDiv())
        elif ctx.MOD():
            return self._visit_bin_op(ctx, ast.Mod())
        elif ctx.AT():
            return self._visit_bin_op(ctx, ast.MatMult())
        return self.visitFactor(ctx.factor())

    def visitFactor(self, ctx: FandangoParser.FactorContext):
        if ctx.ADD():
            return self._visit_unary_op(ctx, ast.UAdd())
        elif ctx.MINUS():
            return self._visit_unary_op(ctx, ast.USub())
        elif ctx.NOT_OP():
            return self._visit_unary_op(ctx, ast.Invert())
        return self.visitPower(ctx.power())

    def visitPower(self, ctx: FandangoParser.PowerContext):
        if ctx.factor():
            return self._visit_bin_op(ctx, ast.Pow())
        return self.visitAwait_primary(ctx.await_primary())

    def visitAwait_primary(self, ctx: FandangoParser.Await_primaryContext):
        if ctx.AWAIT():
            tree, searches, search_map = self.visitPrimary(ctx.primary())
            return ast.Await(value=tree), searches, search_map
        return self.visitPrimary(ctx.primary())

    def visitPrimary(self, ctx: FandangoParser.PrimaryContext):
        if ctx.identifier():
            tree, searches, search_map = self.visitPrimary(ctx.primary())
            return (
                ast.Attribute(value=tree, attr=ctx.identifier().getText()),
                searches,
                search_map,
            )
        elif ctx.genexp():
            tree, searches, search_map = self.visitPrimary(ctx.primary())
            gen_tree, gen_searches, gen_search_map = self.visitGenexp(ctx.genexp())
            return (
                ast.Call(func=tree, args=[gen_tree], keywords=[]),
                searches + gen_searches,
                {**search_map, **gen_search_map},
            )
        elif ctx.OPEN_PAREN():
            tree, searches, search_map = self.visitPrimary(ctx.primary())
            args, keywords = list(), list()
            if ctx.arguments():
                args_trees, args_searches, args_search_map = self.visitArguments(
                    ctx.arguments()
                )
                searches.extend(args_searches)
                search_map.update(args_search_map)
                for arg in args_trees:
                    if isinstance(arg, ast.keyword):
                        keywords.append(arg)
                    else:
                        args.append(arg)
            return (
                ast.Call(func=tree, args=args, keywords=keywords),
                searches,
                search_map,
            )
        elif ctx.slices():
            tree, searches, search_map = self.visitPrimary(ctx.primary())
            return self._process_slices(ctx.slices(), tree, searches, search_map)
        elif ctx.atom():
            return self.visitAtom(ctx.atom())
        else:
            raise FandangoValueError(f"Unsupported atom {ctx.getText()}")

    def _process_slices(self, slices, tree, searches, search_map):
        slice_trees, slice_searches, slice_search_map = self.visitSlices(slices)
        if isinstance(slice_trees, list):
            if len(slice_trees) == 1:
                slice_trees = slice_trees[0]
            else:
                slice_trees = ast.Tuple(elts=slice_trees)
        return (
            ast.Subscript(value=tree, slice=slice_trees),
            searches + slice_searches,
            {**search_map, **slice_search_map},
        )

    def visitAtom(self, ctx: FandangoParser.AtomContext):
        if ctx.identifier():
            return ast.Name(id=ctx.identifier().getText()), [], {}
        elif ctx.TRUE():
            return ast.Constant(value=True), [], {}
        elif ctx.FALSE():
            return ast.Constant(value=False), [], {}
        elif ctx.NONE():
            return ast.Constant(value=None), [], {}
        elif ctx.strings():
            return self.visitStrings(ctx.strings())
        elif ctx.NUMBER():
            return ast.parse(ctx.NUMBER().getText(), mode="eval").body, [], {}
        elif ctx.tuple_():
            return self.visitTuple(ctx.tuple_())
        elif ctx.group():
            return self.visitGroup(ctx.group())
        elif ctx.genexp():
            return self.visitGenexp(ctx.genexp())
        elif ctx.list_():
            return self.visitList(ctx.list_())
        elif ctx.listcomp():
            return self.visitListcomp(ctx.listcomp())
        elif ctx.dict_():
            return self.visitDict(ctx.dict_())
        elif ctx.dictcomp():
            return self.visitDictcomp(ctx.dictcomp())
        elif ctx.set_():
            return self.visitSet(ctx.set_())
        elif ctx.setcomp():
            return self.visitSetcomp(ctx.setcomp())
        elif ctx.ELLIPSIS():
            return ast.Constant(value=Ellipsis), [], {}
        elif ctx.selector_length():
            return self.visitSelector_length(ctx.selector_length())
        else:
            raise FandangoValueError(f"Unsupported atom {ctx.getText()}")

    def visitKwarg_or_starred(self, ctx: FandangoParser.Kwarg_or_starredContext):
        if ctx.ASSIGN():
            tree, searches, search_map = self.visitExpression(ctx.expression())
            return (
                ast.keyword(arg=ctx.identifier().getText(), value=tree),
                searches,
                search_map,
            )
        else:
            return self.visitStarred_expression(ctx.starred_expression())

    def visitKwarg_or_double_starred(
        self, ctx: FandangoParser.Kwarg_or_double_starredContext
    ):
        tree, searches, search_map = self.visitExpression(ctx.expression())
        arg = None
        if ctx.ASSIGN():
            arg = ctx.identifier().getText()
        return (
            ast.keyword(arg=arg, value=tree),
            searches,
            search_map,
        )

    def visitStarred_expression(self, ctx: FandangoParser.Starred_expressionContext):
        tree, searches, search_map = self.visitExpression(ctx.expression())
        return ast.Starred(value=tree), searches, search_map

    def visitAssignment_expression(
        self, ctx: FandangoParser.Assignment_expressionContext
    ):
        raise UnsupportedOperation(
            f"Assignment expressions are currently not supported: {ctx.getText()}"
        )

    def visitSlice(self, ctx: FandangoParser.SliceContext):
        if ctx.COLON():
            slices = [None, None, None]
            slice_count = 0
            searches = list()
            search_map = dict()
            for child in ctx.getChildren():
                if child.getText() == ":":
                    slice_count += 1
                else:
                    slice_tree, slice_searches, slice_search_map = self.visit(child)
                    slices[slice_count] = slice_tree
                    searches.extend(slice_searches)
                    search_map.update(slice_search_map)
            lower, upper, step = slices
            return (
                ast.Slice(
                    lower=lower,
                    upper=upper,
                    step=step,
                    lineno=0,
                    end_lineno=0,
                    col_offset=0,
                    end_col_offset=0,
                ),
                searches,
                search_map,
            )
        else:
            return self.visitNamed_expression(ctx.named_expression())

    def visitStrings(self, ctx: FandangoParser.StringsContext):
        trees, searches, search_map = self.visitChildren(ctx)
        if len(trees) == 1 and isinstance(trees[0], ast.Constant):
            return trees[0], searches, search_map
        return ast.JoinedStr(values=trees), searches, search_map

    def visitString(self, ctx: FandangoParser.StringContext):
        if ctx.STRING():
            value = Terminal.clean(ctx.STRING().getText())
        else:
            raise FandangoValueError(f"Unsupported string: {ctx.getText()}")
        return ast.Constant(value=value), [], {}

    def visitFstring(self, ctx: FandangoParser.FstringContext):
        trees: list[ast.expr]
        searches: list[AttributeSearch]
        search_map: dict[str, AttributeSearch]
        trees, searches, search_map = [], [], {}
        if ctx.FSTRING_START_QUOTE():
            for child in ctx.fstring_middle_no_quote():
                trees, searches, search_map = self.aggregateResult(
                    (trees, searches, search_map),
                    self.visitFstring_middle_no_quote(child),
                )
        elif ctx.FSTRING_START_SINGLE_QUOTE():
            for child in ctx.fstring_middle_no_single_quote():
                trees, searches, search_map = self.aggregateResult(
                    (trees, searches, search_map),
                    self.visitFstring_middle_no_single_quote(child),
                )
        elif ctx.FSTRING_START_TRIPLE_QUOTE():
            for child in ctx.fstring_middle_breaks_no_triple_quote():
                trees, searches, search_map = self.aggregateResult(
                    (trees, searches, search_map),
                    self.visitFstring_middle_breaks_no_triple_quote(child),
                )
        elif ctx.FSTRING_START_TRIPLE_SINGLE_QUOTE():
            for child in ctx.fstring_middle_breaks_no_triple_single_quote():
                trees, searches, search_map = self.aggregateResult(
                    (trees, searches, search_map),
                    self.visitFstring_middle_breaks_no_triple_single_quote(child),
                )
        else:
            raise FandangoValueError(f"Unsupported f-string: {ctx.getText()}")
        # noinspection PyUnreachableCode
        return ast.JoinedStr(values=trees), searches, search_map

    def visitFstring_middle_no_quote(
        self, ctx: FandangoParser.Fstring_middle_no_quoteContext
    ):
        if ctx.fstring_replacement_field():
            return self.visitFstring_replacement_field(ctx.fstring_replacement_field())
        else:
            return ast.Constant(value=ctx.getText()), [], {}

    def visitFstring_middle_no_single_quote(
        self, ctx: FandangoParser.Fstring_middle_no_single_quoteContext
    ):
        if ctx.fstring_replacement_field():
            return self.visitFstring_replacement_field(ctx.fstring_replacement_field())
        else:
            return ast.Constant(value=ctx.getText()), [], {}

    def visitFstring_middle_breaks_no_triple_quote(
        self, ctx: FandangoParser.Fstring_middle_breaks_no_triple_quoteContext
    ):
        if ctx.fstring_replacement_field():
            return self.visitFstring_replacement_field(ctx.fstring_replacement_field())
        else:
            return ast.Constant(value=ctx.getText()), [], {}

    def visitFstring_middle_breaks_no_triple_single_quote(
        self, ctx: FandangoParser.Fstring_middle_breaks_no_triple_single_quoteContext
    ):
        if ctx.fstring_replacement_field():
            return self.visitFstring_replacement_field(ctx.fstring_replacement_field())
        else:
            return ast.Constant(value=ctx.getText()), [], {}

    def visitFstring_replacement_field(
        self, ctx: FandangoParser.Fstring_replacement_fieldContext
    ):
        if ctx.yield_expr():
            tree, searches, search_map = self.visitYield_expr(ctx.yield_expr())
        elif ctx.star_expressions():
            tree, searches, search_map = self.visitStar_expressions(
                ctx.star_expressions()
            )
        else:
            raise FandangoValueError(f"Unsupported f-string: {ctx.getText()}")
        if ctx.fstring_conversion():
            conversion = ord(ctx.fstring_conversion().identifier().getText())
        else:
            conversion = -1
        if ctx.fstring_full_format_spec():
            format_spec, spec_searches, spec_search_map = (
                self.visitFstring_full_format_spec(ctx.fstring_full_format_spec())
            )
            searches.extend(spec_searches)
            search_map.update(spec_search_map)
        else:
            format_spec = None
        return (
            ast.FormattedValue(
                value=tree,
                conversion=conversion,
                format_spec=format_spec,
            ),
            searches,
            search_map,
        )

    def visitFstring_full_format_spec(
        self, ctx: FandangoParser.Fstring_full_format_specContext
    ):
        trees: list[ast.expr] = []
        searches: list[AttributeSearch] = []
        search_map: dict[str, AttributeSearch] = {}
        trees, searches, search_map = [], [], {}
        for child in ctx.fstring_format_spec():
            trees, searches, search_map = self.aggregateResult(
                (trees, searches, search_map),
                self.visitFstring_format_spec(child),
            )
        return ast.JoinedStr(values=trees), searches, search_map

    def visitFstring_format_spec(self, ctx: FandangoParser.Fstring_format_specContext):
        if ctx.fstring_replacement_field():
            return self.visitFstring_replacement_field(ctx.fstring_replacement_field())
        elif ctx.fstring_middle():
            return ast.Constant(value=ctx.getText()), [], {}
        else:
            raise FandangoValueError(
                f"Unsupported f-string format spec: {ctx.getText()}"
            )

    def visitTuple(self, ctx: FandangoParser.TupleContext):
        trees, searches, search_map = self.visitChildren(ctx)
        return ast.Tuple(elts=trees), searches, search_map

    def visitList(self, ctx: FandangoParser.ListContext):
        trees, searches, search_map = self.visitChildren(ctx)
        return ast.List(elts=trees), searches, search_map

    def visitSet(self, ctx: FandangoParser.SetContext):
        trees, searches, search_map = self.visitChildren(ctx)
        return ast.Set(elts=trees), searches, search_map

    def visitDict(self, ctx: FandangoParser.DictContext):
        keys: list[ast.expr | None] = list()
        values: list[ast.expr] = list()
        if ctx.double_starred_kvpairs():
            kvpairs, searches, search_map = self.visitDouble_starred_kvpairs(
                ctx.double_starred_kvpairs()
            )
            keys_, values_ = zip(*kvpairs)
            keys.extend(keys_)
            values.extend(values_)
        else:
            searches, search_map = [], {}
        return ast.Dict(keys=keys, values=values), searches, search_map

    def visitDouble_starred_kvpair(
        self, ctx: FandangoParser.Double_starred_kvpairContext
    ):
        if ctx.POWER():
            tree, searches, search_map = self.visitBitwise_or(ctx.bitwise_or())
            return (None, tree), searches, search_map
        else:
            return self.visitKvpair(ctx.kvpair())

    def visitKvpair(self, ctx: FandangoParser.KvpairContext):
        key, key_searches, key_search_map = self.visitExpression(ctx.expression(0))
        value, value_searches, value_search_map = self.visitExpression(
            ctx.expression(1)
        )
        return (
            (key, value),
            key_searches + value_searches,
            {**key_search_map, **value_search_map},
        )

    def visitGroup(self, ctx: FandangoParser.GroupContext):
        if ctx.yield_expr():
            return self.visitYield_expr(ctx.yield_expr())
        else:
            return self.visitNamed_expression(ctx.named_expression())

    def visitYield_expr(self, ctx: FandangoParser.Yield_exprContext):
        if ctx.FROM():
            tree, searches, search_map = self.visitExpression(ctx.expression())
            return ast.YieldFrom(value=tree), searches, search_map
        else:
            if ctx.star_expressions():
                tree, searches, search_map = self.visitStar_expressions(
                    ctx.star_expressions()
                )
                return ast.Yield(value=tree), searches, search_map
            else:
                return ast.Yield(value=None), [], {}

    def visitNamed_expression(self, ctx: FandangoParser.Named_expressionContext):
        if ctx.expression():
            return self.visitExpression(ctx.expression())
        else:
            return self.visitAssignment_expression(ctx.assignment_expression())

    def visitGenexp(self, ctx: FandangoParser.GenexpContext):
        if ctx.expression():
            tree, searches, search_map = self.visitExpression(ctx.expression())
        else:
            tree, searches, search_map = self.visitAssignment_expression(
                ctx.assignment_expression()
            )
        generators, gen_searches, gen_search_map = self.visitFor_if_clauses(
            ctx.for_if_clauses()
        )
        return (
            ast.GeneratorExp(elt=tree, generators=generators),
            searches + gen_searches,
            {**search_map, **gen_search_map},
        )

    def visitListcomp(self, ctx: FandangoParser.ListcompContext):
        tree, searches, search_map = self.visitNamed_expression(ctx.named_expression())
        generators, gen_searches, gen_search_map = self.visitFor_if_clauses(
            ctx.for_if_clauses()
        )
        return (
            ast.ListComp(elt=tree, generators=generators),
            searches + gen_searches,
            {**search_map, **gen_search_map},
        )

    def visitSetcomp(self, ctx: FandangoParser.SetcompContext):
        tree, searches, search_map = self.visitNamed_expression(ctx.named_expression())
        generators, gen_searches, gen_search_map = self.visitFor_if_clauses(
            ctx.for_if_clauses()
        )
        return (
            ast.SetComp(elt=tree, generators=generators),
            searches + gen_searches,
            {**search_map, **gen_search_map},
        )

    def visitDictcomp(self, ctx: FandangoParser.DictcompContext):
        tree, searches, search_map = self.visitKvpair(ctx.kvpair())
        key, value = tree
        generators, gen_searches, gen_search_map = self.visitFor_if_clauses(
            ctx.for_if_clauses()
        )
        return (
            ast.DictComp(key=key, value=value, generators=generators),
            searches + gen_searches,
            {**search_map, **gen_search_map},
        )

    def visitArguments(self, ctx: FandangoParser.ArgumentsContext):
        return self.visitArgs(ctx.args())

    def visitArgs(self, ctx: FandangoParser.ArgsContext):
        if ctx.kwargs() and not ctx.arg():
            return self.visitKwargs(ctx.kwargs())

        result: tuple[list, Any, dict] = list(), list(), dict()
        for arg in ctx.arg():
            result = self.aggregateResult(result, self.visitArg(arg))
        if ctx.kwargs():
            result = self.aggregateResult(result, self.visitKwargs(ctx.kwargs()))
        return result

    def visitArg(self, ctx: FandangoParser.ArgContext):
        if ctx.star_selection():
            return self.visitStar_selection(ctx.star_selection())
        elif ctx.starred_expression():
            return self.visitStarred_expression(ctx.starred_expression())
        elif ctx.assignment_expression():
            return self.visitAssignment_expression(ctx.assignment_expression())
        elif ctx.expression():
            return self.visitExpression(ctx.expression())
        else:
            raise FandangoValueError(f"Unknown argument: {ctx.getText()}")

    def visitKwargs(self, ctx: FandangoParser.KwargsContext):
        result: tuple[list, Any, dict] = list(), list(), dict()

        for kwarg in ctx.kwarg_or_starred():
            result = self.aggregateResult(result, self.visitKwarg_or_starred(kwarg))
        for kwarg in ctx.kwarg_or_double_starred():
            result = self.aggregateResult(
                result, self.visitKwarg_or_double_starred(kwarg)
            )

        return result

    def visitFor_if_clauses(self, ctx: FandangoParser.For_if_clausesContext):
        result: tuple[list, Any, dict] = list(), list(), dict()
        for clause in ctx.for_if_clause():
            result = self.aggregateResult(result, self.visitFor_if_clause(clause))
        return result

    def visitFor_if_clause(self, ctx: FandangoParser.For_if_clauseContext):
        is_async = True if ctx.ASYNC() else False  # needed for None check
        target, target_searches, target_search_map = self.visitStar_targets(
            ctx.star_targets()
        )
        iter_, iter_searches, iter_search_map = self.visitDisjunction(
            ctx.disjunction(0)
        )
        searches = target_searches + iter_searches
        search_map = {**target_search_map, **iter_search_map}
        ifs = list()
        for disjunction in ctx.disjunction()[1:]:
            if_, if_searches, if_search_map = self.visitDisjunction(disjunction)
            ifs.append(if_)
            searches.extend(if_searches)
            search_map.update(if_search_map)
        return (
            ast.comprehension(
                target=target,
                iter=iter_,
                ifs=ifs,
                is_async=is_async,
            ),
            searches,
            search_map,
        )

    def visitStar_targets(self, ctx: FandangoParser.Star_targetsContext):
        targets, searches, search_map = self.visitChildren(ctx)
        if isinstance(targets, list):
            if len(targets) == 1:
                target = targets[0]
            else:
                target = ast.Tuple(elts=targets)
        else:
            target = targets
        return target, searches, search_map

    def visitStar_targets_list_seq(
        self, ctx: FandangoParser.Star_targets_list_seqContext
    ):
        targets, searches, search_map = self.visitChildren(ctx)
        if not isinstance(targets, list):
            targets = [targets]
        return ast.List(
            elts=targets,
        )

    def visitStar_targets_tuple_seq(
        self, ctx: FandangoParser.Star_targets_tuple_seqContext
    ):
        targets, searches, search_map = self.visitChildren(ctx)
        if not isinstance(targets, list):
            targets = [targets]
        return ast.Tuple(
            elts=targets,
        )

    def visitStar_target(self, ctx: FandangoParser.Star_targetContext):
        if ctx.STAR():
            target, searches, search_map = self.visitStar_target(ctx.star_target())
            return ast.Starred(value=target), searches, search_map
        return self.visitTarget_with_star_atom(ctx.target_with_star_atom())

    def visitTarget_with_star_atom(
        self, ctx: FandangoParser.Target_with_star_atomContext
    ):
        if ctx.DOT():
            target, searches, search_map = self.visitT_primary(ctx.t_primary())
            return (
                ast.Attribute(value=target, attr=ctx.identifier().getText()),
                searches,
                search_map,
            )
        elif ctx.slices():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            return self._process_slices(ctx.slices(), tree, searches, search_map)
        return self.visitStar_atom(ctx.star_atom())

    def visitStar_atom(self, ctx: FandangoParser.Star_atomContext):
        if ctx.identifier():
            return ast.Name(id=ctx.identifier().getText()), [], {}
        elif ctx.target_with_star_atom():
            return self.visitTarget_with_star_atom(ctx.target_with_star_atom())
        elif ctx.OPEN_PAREN():
            if ctx.star_targets_tuple_seq():
                return self.visitStar_targets_tuple_seq(ctx.star_targets_tuple_seq())
            return ast.Tuple(elts=[]), [], {}
        elif ctx.OPEN_BRACK():
            if ctx.star_targets_list_seq():
                return self.visitStar_targets_list_seq(ctx.star_targets_list_seq())
            return ast.List(elts=[]), [], {}
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitT_primary(self, ctx: FandangoParser.T_primaryContext):
        if ctx.DOT():
            target, searches, search_map = self.visitT_primary(ctx.t_primary())
            return (
                ast.Attribute(value=target, attr=ctx.identifier().getText()),
                searches,
                search_map,
            )
        elif ctx.slices():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            return self._process_slices(ctx.slices(), tree, searches, search_map)
        elif ctx.genexp():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            gen_tree, gen_searches, gen_search_map = self.visitGenexp(ctx.genexp())
            return (
                ast.Call(func=tree, args=[gen_tree], keywords=[]),
                searches + gen_searches,
                {**search_map, **gen_search_map},
            )
        elif ctx.OPEN_PAREN():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            args, keywords = list(), list()
            if ctx.arguments():
                args_trees, args_searches, args_search_map = self.visitArguments(
                    ctx.arguments()
                )
                searches.extend(args_searches)
                search_map.update(args_search_map)
                for arg in args_trees:
                    if isinstance(arg, ast.keyword):
                        keywords.append(arg)
                    else:
                        args.append(arg)
            return (
                ast.Call(func=tree, args=args, keywords=keywords),
                searches,
                search_map,
            )
        return self.visitAtom(ctx.atom())

    def visitParameters(self, ctx: FandangoParser.ParametersContext):
        posonlyargs, defaults = [], []
        searches = []
        search_map = {}
        if ctx.slash_no_default():
            p = self.visitSlash_no_default(ctx.slash_no_default())
            posonlyargs.extend(p)
        elif ctx.slash_with_default():
            p, d, s, m = self.visitSlash_with_default(ctx.slash_with_default())
            posonlyargs.extend(p)
            defaults.extend(d)
            searches.extend(s)
            search_map.update(m)
        args = []
        for param in ctx.param_no_default():
            arg, s, m = self.visitParam_no_default(param)
            args.append(arg)
            searches.extend(s)
            search_map.update(m)
        kwonlyargs = []
        kw_defaults = []
        for param in ctx.param_with_default():
            arg, d, s, m = self.visitParam_with_default(param)
            if ctx.slash_with_default():
                args.append(arg)
                defaults.append(d)
            else:
                kwonlyargs.append(arg)
                kw_defaults.append(d)
            searches.extend(s)
            search_map.update(m)
        if ctx.star_etc():
            vararg, kw_args, kw_d, kwarg, s, m = self.visitStar_etc(ctx.star_etc())
            kwonlyargs.extend(kw_args)
            kw_defaults.extend(kw_d)
            searches.extend(s)
            search_map.update(m)
        else:
            vararg, kwarg = None, None
        return (
            ast.arguments(
                posonlyargs=posonlyargs,
                args=args,
                vararg=vararg,
                kwonlyargs=kwonlyargs,
                kw_defaults=kw_defaults,
                kwarg=kwarg,
                defaults=defaults,
            ),
            searches,
            search_map,
        )

    def visitStar_etc(self, ctx: FandangoParser.Star_etcContext):
        searches, search_map = [], {}
        vararg = None
        if ctx.STAR():
            if ctx.param_no_default():
                vararg, s, m = self.visitParam_no_default(ctx.param_no_default())
            else:
                vararg, s, m = self.visitParam_no_default_star_annotation(
                    ctx.param_no_default_star_annotation()
                )
            searches.extend(s)
            search_map.update(m)
        kwonlyargs = []
        kw_defaults = []
        for param in ctx.param_maybe_default():
            arg, d, s, m = self.visitParam_maybe_default(param)
            kwonlyargs.append(arg)
            kw_defaults.append(d)
            searches.extend(s)
            search_map.update(m)
        kwarg = None
        if ctx.kwds():
            kwarg, s, m = self.visitParam_no_default(ctx.kwds().param_no_default())
            searches.extend(s)
            search_map.update(m)
        return vararg, kwonlyargs, kw_defaults, kwarg, searches, search_map

    def visitParam_no_default(self, ctx: FandangoParser.Param_no_defaultContext):
        return self.visitParam(ctx.param())

    def visitParam_no_default_star_annotation(
        self, ctx: FandangoParser.Param_no_default_star_annotationContext
    ):
        return self.visitParam_star_annotation(ctx.param_star_annotation())

    def visitParam_with_default(self, ctx: FandangoParser.Param_with_defaultContext):
        arg, searches, search_map = self.visitParam(ctx.param())
        default, default_searches, default_search_map = self.visitExpression(
            ctx.default().expression()
        )
        return (
            arg,
            default,
            searches + default_searches,
            {**search_map, **default_search_map},
        )

    def visitParam_maybe_default(self, ctx: FandangoParser.Param_maybe_defaultContext):
        arg, searches, search_map = self.visitParam(ctx.param())
        if ctx.default():
            default, default_searches, default_search_map = self.visitExpression(
                ctx.default().expression()
            )
            return (
                arg,
                default,
                searches + default_searches,
                {**search_map, **default_search_map},
            )
        else:
            return arg, None, searches, search_map

    def visitParam(self, ctx: FandangoParser.ParamContext):
        if ctx.annotation():
            annotation, searches, search_map = self.visitExpression(
                ctx.annotation().expression()
            )
        else:
            annotation, searches, search_map = None, [], {}
        return (
            ast.arg(
                arg=ctx.identifier().getText(),
                annotation=annotation,
            ),
            searches,
            search_map,
        )

    def visitParam_star_annotation(
        self, ctx: FandangoParser.Param_star_annotationContext
    ):
        if ctx.star_annotation():
            annotation, searches, search_map = self.visitStar_expression(
                ctx.star_annotation().expression()
            )
        else:
            annotation, searches, search_map = None, [], {}
        return (
            ast.arg(
                arg=ctx.identifier().getText(),
                annotation=annotation,
            ),
            searches,
            search_map,
        )

    def visitSingle_target(self, ctx: FandangoParser.Single_targetContext):
        if ctx.identifier():
            return ast.Name(id=ctx.identifier().getText()), [], {}
        elif ctx.single_subscript_attribute_target():
            return self.visitSingle_subscript_attribute_target(
                ctx.single_subscript_attribute_target()
            )
        else:
            return self.visitSingle_target(ctx.single_target())

    def visitSingle_subscript_attribute_target(
        self, ctx: FandangoParser.Single_subscript_attribute_targetContext
    ):
        tree, searches, search_map = self.visitT_primary(ctx.t_primary())
        if ctx.DOT():
            return (
                ast.Attribute(value=tree, attr=ctx.identifier().getText()),
                searches,
                search_map,
            )
        else:
            return self._process_slices(ctx.slices(), tree, searches, search_map)

    def visitDel_target(self, ctx: FandangoParser.Del_targetContext):
        if ctx.DOT():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            return (
                ast.Attribute(value=tree, attr=ctx.identifier().getText()),
                searches,
                search_map,
            )
        elif ctx.slices():
            tree, searches, search_map = self.visitT_primary(ctx.t_primary())
            return self._process_slices(ctx.slices(), tree, searches, search_map)
        else:
            return self.visitDel_t_atom(ctx.del_t_atom())

    def visitDel_t_atom(self, ctx: FandangoParser.Del_t_atomContext):
        if ctx.identifier():
            return ast.Name(id=ctx.identifier().getText()), [], {}
        else:
            if ctx.del_targets():
                trees, searches, search_map = self.visitDel_targets(ctx.del_targets())
            else:
                trees, searches, search_map = self.defaultResult()
            if ctx.OPEN_PAREN():
                return ast.Tuple(elts=trees), searches, search_map
            else:
                return ast.List(elts=trees), searches, search_map

    def visitStar_expressions(self, ctx: FandangoParser.Star_expressionsContext):
        expressions, searches, search_map = self.visitChildren(ctx)
        if isinstance(expressions, list):
            if len(expressions) == 1:
                expression = expressions[0]
            else:
                expression = ast.Tuple(elts=expressions)
        else:
            expression = expressions
        return expression, searches, search_map

    def visitStar_expression(self, ctx: FandangoParser.Star_expressionContext):
        if ctx.STAR():
            tree, searches, search_map = self.visitBitwise_or(ctx.bitwise_or())
            return ast.Starred(value=tree), searches, search_map
        else:
            return self.visitExpression(ctx.expression())


class PythonProcessor(FandangoParserVisitor):
    def __init__(self):
        self.search_processor = SearchProcessor(Grammar.dummy())

    def get_code(self, stmts: list[FandangoParser.PythonContext]):
        return ast.Module(body=[self.visit(stmt) for stmt in stmts], type_ignores=[])

    def get_expression(self, expression):
        if expression:
            tree, searches, _ = self.search_processor.visit(expression)
            if searches:
                raise FandangoValueError(
                    "Nonterminals can only be used in grammars and constraints, "
                    f"not in regular Python code: {expression.getText()}"
                )
            return tree
        return None

    def visitStatements(self, ctx: FandangoParser.StatementsContext):
        return [tree for stmt in ctx.stmt() for tree in self.visitStmt(stmt)]

    def visitSimple_stmts(self, ctx: FandangoParser.Simple_stmtsContext):
        return [self.visitSimple_stmt(stmt) for stmt in ctx.simple_stmt()]

    def visitStmt(self, ctx: FandangoParser.StmtContext):
        if ctx.compound_stmt():
            return [self.visitCompound_stmt(ctx.compound_stmt())]
        elif ctx.simple_stmts():
            return self.visitSimple_stmts(ctx.simple_stmts())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitSimple_stmt(self, ctx: FandangoParser.Simple_stmtContext):
        if ctx.assignment():
            return self.visitAssignment(ctx.assignment())
        elif ctx.type_alias():
            return self.visitType_alias(ctx.type_alias())
        elif ctx.star_expressions():
            return ast.Expr(value=self.get_expression(ctx.star_expressions()))
        elif ctx.return_stmt():
            return self.visitReturn_stmt(ctx.return_stmt())
        elif ctx.import_stmt():
            return self.visitImport_stmt(ctx.import_stmt())
        elif ctx.raise_stmt():
            return self.visitRaise_stmt(ctx.raise_stmt())
        elif ctx.PASS():
            return ast.Pass()
        elif ctx.del_stmt():
            return self.visitDel_stmt(ctx.del_stmt())
        elif ctx.yield_stmt():
            return self.visitYield_stmt(ctx.yield_stmt())
        elif ctx.assert_stmt():
            return self.visitAssert_stmt(ctx.assert_stmt())
        elif ctx.BREAK():
            return ast.Break()
        elif ctx.CONTINUE():
            return ast.Continue()
        elif ctx.global_stmt():
            return self.visitGlobal_stmt(ctx.global_stmt())
        elif ctx.nonlocal_stmt():
            return self.visitNonlocal_stmt(ctx.nonlocal_stmt())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitAssignment(self, ctx: FandangoParser.AssignmentContext):
        if ctx.COLON():
            right = self.get_expression(ctx.annotated_rhs())
            if ctx.identifier():
                return ast.AnnAssign(
                    target=ast.Name(id=ctx.identifier().getText()),
                    annotation=self.get_expression(ctx.expression()),
                    value=right,
                    simple=1,
                )
            if ctx.single_target():
                left = self.get_expression(ctx.single_target())
            elif ctx.single_subscript_attribute_target():
                left = self.get_expression(ctx.single_subscript_attribute_target())
            else:
                raise FandangoValueError(
                    f"Unsupported left hand side for assignment {ctx.getText()}"
                )
            return ast.AnnAssign(
                target=left,
                annotation=self.get_expression(ctx.expression()),
                value=right,
                simple=0,
                lineno=0,
            )
        else:
            if ctx.yield_expr():
                value = self.get_expression(ctx.yield_expr())
            elif ctx.star_expressions():
                value = self.get_expression(ctx.star_expressions())
            else:
                raise FandangoValueError(
                    f"Unsupported right hand side for assignment {ctx.getText()}"
                )
            if ctx.ASSIGN():
                return ast.Assign(
                    targets=[
                        self.get_expression(target) for target in ctx.star_targets()
                    ],
                    value=value,
                    lineno=0,
                )
            aug: FandangoParser.AugassignContext = ctx.augassign()
            op: ast.operator
            if aug.ADD_ASSIGN():
                op = ast.Add()
            elif aug.SUB_ASSIGN():
                op = ast.Sub()
            elif aug.MULT_ASSIGN():
                op = ast.Mult()
            elif aug.AT_ASSIGN():
                op = ast.MatMult()
            elif aug.DIV_ASSIGN():
                op = ast.Div()
            elif aug.MOD_ASSIGN():
                op = ast.Mod()
            elif aug.AND_ASSIGN():
                op = ast.BitAnd()
            elif aug.OR_ASSIGN():
                op = ast.BitOr()
            elif aug.XOR_ASSIGN():
                op = ast.BitXor()
            elif aug.LEFT_SHIFT_ASSIGN():
                op = ast.LShift()
            elif aug.RIGHT_SHIFT_ASSIGN():
                op = ast.RShift()
            elif aug.POWER_ASSIGN():
                op = ast.Pow()
            elif aug.IDIV_ASSIGN():
                op = ast.FloorDiv()
            else:
                raise FandangoValueError(
                    f"Unsupported operator for augmented assignment: {aug.getText()}"
                )
            return ast.AugAssign(
                target=self.get_expression(ctx.single_target()),
                op=op,
                value=value,
                lineno=0,
            )

    def visitType_alias(self, ctx: FandangoParser.Type_aliasContext):
        raise UnsupportedOperation(
            "Type alias currently not supported: {ctx.getText()}"
        )

    def visitReturn_stmt(self, ctx: FandangoParser.Return_stmtContext):
        return ast.Return(value=self.get_expression(ctx.star_expressions()))

    def visitRaise_stmt(self, ctx: FandangoParser.Raise_stmtContext):
        return ast.Raise(
            exc=self.get_expression(ctx.expression(0)),
            cause=self.get_expression(ctx.expression(1)),
        )

    def visitGlobal_stmt(self, ctx: FandangoParser.Global_stmtContext):
        return ast.Global(names=[name.getText() for name in ctx.identifier()])

    def visitNonlocal_stmt(self, ctx: FandangoParser.Nonlocal_stmtContext):
        return ast.Nonlocal(names=[name.getText() for name in ctx.identifier()])

    def visitDel_stmt(self, ctx: FandangoParser.Del_stmtContext):
        return ast.Delete(targets=self.get_expression(ctx.del_targets()))

    def visitYield_stmt(self, ctx: FandangoParser.Yield_stmtContext):
        return ast.Expr(value=self.get_expression(ctx.yield_expr()))

    def visitAssert_stmt(self, ctx: FandangoParser.Assert_stmtContext):
        return ast.Assert(
            test=self.get_expression(ctx.expression(0)),
            msg=self.get_expression(ctx.expression(1)),
        )

    def visitImport_stmt(self, ctx: FandangoParser.Import_stmtContext):
        if ctx.import_name():
            return self.visitImport_name(ctx.import_name())
        elif ctx.import_from():
            return self.visitImport_from(ctx.import_from())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitImport_name(self, ctx: FandangoParser.Import_nameContext):
        return ast.Import(names=self.visitDotted_as_names(ctx.dotted_as_names()))

    def visitImport_from(self, ctx: FandangoParser.Import_fromContext):
        level = len(ctx.DOT()) + 3 * len(ctx.ELLIPSIS())
        return ast.ImportFrom(
            module=ctx.dotted_name().getText() if ctx.dotted_name() else None,
            names=self.visitImport_from_targets(ctx.import_from_targets()),
            level=level,
        )

    def visitDotted_as_names(self, ctx: FandangoParser.Dotted_as_namesContext):
        return [self.visitDotted_as_name(name) for name in ctx.dotted_as_name()]

    def visitDotted_as_name(self, ctx: FandangoParser.Dotted_as_nameContext):
        return ast.alias(
            name=ctx.dotted_name().getText(),
            asname=ctx.identifier().getText() if ctx.identifier() else None,
        )

    def visitImport_from_targets(self, ctx: FandangoParser.Import_from_targetsContext):
        if ctx.STAR():
            return [ast.alias(name="*", asname=None)]
        else:
            return self.visitImport_from_as_names(ctx.import_from_as_names())

    def visitImport_from_as_names(
        self, ctx: FandangoParser.Import_from_as_namesContext
    ):
        return [
            self.visitImport_from_as_name(name) for name in ctx.import_from_as_name()
        ]

    def visitImport_from_as_name(self, ctx: FandangoParser.Import_from_as_nameContext):
        return ast.alias(
            name=ctx.identifier(0).getText(),
            asname=ctx.identifier(1).getText() if ctx.identifier(1) else None,
        )

    def visitCompound_stmt(self, ctx: FandangoParser.Compound_stmtContext):
        if ctx.function_def():
            return self.visitFunction_def(ctx.function_def())
        elif ctx.if_stmt():
            return self.visitIf_stmt(ctx.if_stmt())
        elif ctx.class_def():
            return self.visitClass_def(ctx.class_def())
        elif ctx.with_stmt():
            return self.visitWith_stmt(ctx.with_stmt())
        elif ctx.for_stmt():
            return self.visitFor_stmt(ctx.for_stmt())
        elif ctx.try_stmt():
            return self.visitTry_stmt(ctx.try_stmt())
        elif ctx.while_stmt():
            return self.visitWhile_stmt(ctx.while_stmt())
        elif ctx.match_stmt():
            return self.visitMatch_stmt(ctx.match_stmt())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitClass_def(self, ctx: FandangoParser.Class_defContext):
        class_def = self.visitClass_def_raw(ctx.class_def_raw())
        if ctx.decorators():
            class_def.decorator_list = self.visitDecorators(ctx.decorators())
        return class_def

    def visitClass_def_raw(self, ctx: FandangoParser.Class_def_rawContext):
        bases = list()
        keywords = list()
        if ctx.type_params():
            raise UnsupportedOperation(
                f"Type params unsupported for class definitions {ctx.getText()}"
            )
        if ctx.arguments():
            base_trees, base_searches, _ = self.search_processor.visitArguments(
                ctx.arguments()
            )
            if base_searches:
                raise FandangoValueError(
                    "Nonterminals can only be used in grammars and constraints, "
                    f"not in regular Python code: {ctx.getText()}"
                )
            for base in base_trees:
                if isinstance(base, ast.keyword):
                    keywords.append(base)
                else:
                    bases.append(base)
        body = self.visitBlock(ctx.block())
        return ast.ClassDef(
            name=ctx.identifier().getText(),
            bases=bases,
            keywords=keywords,
            body=body,
            decorator_list=[],
        )

    def visitFunction_def(self, ctx: FandangoParser.Function_defContext):
        function_def = self.visitFunction_def_raw(ctx.function_def_raw())
        if ctx.decorators():
            function_def.decorator_list = self.visitDecorators(ctx.decorators())
        return function_def

    def visitFunction_def_raw(self, ctx: FandangoParser.Function_def_rawContext):
        if ctx.type_params():
            raise UnsupportedOperation(
                f"Type params unsupported for class definitions: {ctx.getText()}"
            )
        if ctx.params():
            params = self.visitParams(ctx.params())
        else:
            params = list()
        body = self.visitBlock(ctx.block())
        class_: type[ast.FunctionDef | ast.AsyncFunctionDef]
        if ctx.ASYNC():
            class_ = ast.AsyncFunctionDef
        else:
            class_ = ast.FunctionDef
        return class_(
            name=ctx.identifier().getText(),
            args=params,
            body=body,
            decorator_list=[],
            returns=self.get_expression(ctx.expression()),
            type_comment=None,
            lineno=0,
            end_lineno=0,
            col_offset=0,
            end_col_offset=0,
        )

    def visitDecorators(self, ctx: FandangoParser.DecoratorsContext):
        decorators = list()
        for expression in ctx.named_expression():
            tree, searches, _ = self.search_processor.visitNamed_expression(expression)
            if searches:
                raise FandangoValueError(
                    "Nonterminals can only be used in grammars and constraints, "
                    f"not in regular Python code: {ctx.getText()}"
                )
            decorators.append(tree)
        return decorators

    def visitBlock(self, ctx: FandangoParser.BlockContext):
        if ctx.statements():
            return self.visitStatements(ctx.statements())
        elif ctx.simple_stmts():
            return self.visitSimple_stmts(ctx.simple_stmts())
        else:
            raise FandangoValueError(f"Unknown symbol: {ctx.getText()}")

    def visitParams(self, ctx: FandangoParser.ParamsContext):
        return self.get_expression(ctx.parameters())

    def _process_if(
        self, ctx: FandangoParser.If_stmtContext | FandangoParser.Elif_stmtContext
    ):
        test = self.get_expression(ctx.named_expression())
        body = self.visitBlock(ctx.block())
        if ctx.elif_stmt():
            orelse = [self._process_if(ctx.elif_stmt())]
        elif ctx.else_block():
            orelse = self.visitBlock(ctx.else_block().block())
        else:
            orelse = []
        return ast.If(
            test=test,
            body=body,
            orelse=orelse,
        )

    def visitIf_stmt(self, ctx: FandangoParser.If_stmtContext):
        return self._process_if(ctx)

    def visitElif_stmt(self, ctx: FandangoParser.Elif_stmtContext):
        return self._process_if(ctx)

    def visitWith_stmt(self, ctx: FandangoParser.With_stmtContext):
        items = [self.visitWith_item(item) for item in ctx.with_item()]
        body = self.visitBlock(ctx.block())
        if ctx.ASYNC():
            return ast.AsyncWith(
                items=items,
                body=body,
                lineno=0,
            )
        return ast.With(
            items=items,
            body=body,
            lineno=0,
        )

    def visitWith_item(self, ctx: FandangoParser.With_itemContext):
        return ast.withitem(
            context_expr=self.get_expression(ctx.expression()),
            optional_vars=self.get_expression(ctx.star_target()),
        )

    def visitFor_stmt(self, ctx: FandangoParser.For_stmtContext):
        target = self.get_expression(ctx.star_targets())
        iter_ = self.get_expression(ctx.star_expressions())
        body = self.visitBlock(ctx.block())
        if ctx.else_block():
            orelse = self.visitBlock(ctx.else_block().block())
        else:
            orelse = None
        if ctx.ASYNC():
            return ast.AsyncFor(
                target=target,
                iter=iter_,
                body=body,
                orelse=orelse,
                lineno=0,
            )
        return ast.For(
            target=target,
            iter=iter_,
            body=body,
            orelse=orelse,
            lineno=0,
        )

    def visitWhile_stmt(self, ctx: FandangoParser.While_stmtContext):
        if ctx.else_block():
            orelse = self.visitBlock(ctx.else_block().block())
        else:
            orelse = None
        return ast.While(
            test=self.get_expression(ctx.named_expression()),
            body=self.visitBlock(ctx.block()),
            orelse=orelse,
        )

    def visitTry_stmt(self, ctx: FandangoParser.Try_stmtContext):
        body = self.visitBlock(ctx.block())
        if ctx.else_block():
            orelse = self.visitBlock(ctx.else_block().block())
        else:
            orelse = None
        if ctx.finally_block():
            finalbody = self.visitBlock(ctx.finally_block().block())
        else:
            finalbody = None
        if ctx.except_star_block():
            return ast.TryStar(
                body=body,
                handlers=[
                    self.visitExcept_star_block(handler)
                    for handler in ctx.except_star_block()
                ],
                orelse=orelse,
                finalbody=finalbody,
            )
        return ast.Try(
            body=body,
            handlers=[
                self.visitExcept_block(handler) for handler in ctx.except_block()
            ],
            orelse=orelse,
            finalbody=finalbody,
        )

    def visitExcept_block(self, ctx: FandangoParser.Except_blockContext):
        if ctx.identifier():
            name = ctx.identifier().getText()
        else:
            name = None
        return ast.ExceptHandler(
            type=self.get_expression(ctx.expression()),
            name=name,
            body=self.visitBlock(ctx.block()),
        )

    def visitExcept_star_block(self, ctx: FandangoParser.Except_star_blockContext):
        if ctx.identifier():
            name = ctx.identifier().getText()
        else:
            name = None
        return ast.ExceptHandler(
            type=self.get_expression(ctx.expression()),
            name=name,
            body=self.visitBlock(ctx.block()),
        )

    def visitMatch_stmt(self, ctx: FandangoParser.Match_stmtContext):
        raise UnsupportedOperation(
            f"Match statement currently not supported: {ctx.getText()}"
        )



import argparse
from collections.abc import Callable
import glob
import os
from pathlib import Path
import shutil
from ansi_styles import ansiStyles as styles
import sys
import tempfile
from typing import Any

import fandango
from fandango.api import Fandango
from fandango.cli.output import open_file, output_population, output_solution
from fandango.cli.parser import get_parser
from fandango.cli.utils import (
    get_file_mode,
    make_fandango_settings,
    parse_constraints_from_args,
    parse_contents_from_args,
    parse_file,
    validate,
)
from fandango.constraints.constraint import Constraint
from fandango.constraints.soft import SoftValue
from fandango.converters.FandangoConverter import FandangoConverter
from fandango.converters.antlr.ANTLRFandangoConverter import ANTLRFandangoConverter
from fandango.converters.bt.BTFandangoConverter import (
    BitfieldOrder,
    BTFandangoConverter,
    Endianness,
)
from fandango.converters.dtd.DTDFandangoConverter import DTDFandangoConverter
from fandango.converters.fan.FandangoFandangoConverter import FandangoFandangoConverter
from fandango.errors import FandangoError, FandangoParseError
from fandango.language.grammar import FuzzingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.parse import cache_dir, clear_cache
from fandango.logger import LOGGER, print_exception


def help_command(args: argparse.Namespace, in_command_line: bool = True) -> None:
    parser = get_parser(in_command_line)

    help_issued = False
    for cmd in args.help_command:
        try:
            parser.parse_args([cmd] + ["--help"])
            help_issued = True
        except SystemExit:
            help_issued = True
            pass
        except argparse.ArgumentError:
            print("Unknown command:", cmd, file=sys.stderr)

    if not help_issued:
        parser.print_help()


def exit_command(args: argparse.Namespace) -> None:
    pass


# Default Fandango file content (grammar, constraints); set with `set`
DEFAULT_FAN_CONTENT: tuple[Grammar | None, list[Constraint | SoftValue]] = (None, [])

# Additional Fandango constraints; set with `set`
DEFAULT_CONSTRAINTS: list[Constraint | SoftValue] = []

# Default Fandango algorithm settings; set with `set`
DEFAULT_SETTINGS: dict[str, Any] = {}


def set_command(args: argparse.Namespace) -> None:
    """Set global settings"""
    global DEFAULT_FAN_CONTENT
    global DEFAULT_CONSTRAINTS
    global DEFAULT_SETTINGS

    if args.fan_files:
        LOGGER.info("Parsing Fandango content")
        grammar, constraints = parse_contents_from_args(args)
        DEFAULT_FAN_CONTENT = (grammar, constraints)
        DEFAULT_CONSTRAINTS = []  # Don't leave these over
    elif args.constraints or args.maxconstraints or args.minconstraints:
        default_grammar = DEFAULT_FAN_CONTENT[0]
        if not default_grammar:
            raise FandangoError("Open a `.fan` file first ('set -f FILE.fan')")

        LOGGER.info("Parsing Fandango constraints")
        _, constraints = parse_constraints_from_args(
            args, given_grammars=[default_grammar]
        )
        DEFAULT_CONSTRAINTS = constraints

    settings = make_fandango_settings(args)
    for setting in settings:
        DEFAULT_SETTINGS[setting] = settings[setting]

    no_args = not args.fan_files and not args.constraints and not settings

    if no_args:
        # Report current settings
        LOGGER.info("Did not receive an arg for set, printing settings")
        grammar, constraints = DEFAULT_FAN_CONTENT
        if grammar:
            for symbol in grammar.rules:
                print(grammar.get_repr_for_rule(symbol))
        if constraints:
            for constraint in constraints:
                print("where " + str(constraint))

    if no_args or (DEFAULT_CONSTRAINTS and sys.stdin.isatty()):
        for constraint in DEFAULT_CONSTRAINTS:
            print("where " + str(constraint) + "  # set by user")
    if no_args or (DEFAULT_SETTINGS and sys.stdin.isatty()):
        for setting in DEFAULT_SETTINGS:
            print(
                "--" + setting.replace("_", "-") + "=" + str(DEFAULT_SETTINGS[setting])
            )


def reset_command(args: argparse.Namespace) -> None:
    """Reset global settings"""
    global DEFAULT_SETTINGS
    DEFAULT_SETTINGS = {}

    global DEFAULT_CONSTRAINTS
    DEFAULT_CONSTRAINTS = []


def cd_command(args: argparse.Namespace) -> None:
    """Change current directory"""
    if args.directory:
        os.chdir(args.directory)
    else:
        os.chdir(Path.home())

    if sys.stdin.isatty():
        print(os.getcwd())


def fuzz_command(args: argparse.Namespace) -> None:
    """Invoke the fuzzer"""

    LOGGER.info("---------- Parsing FANDANGO content ----------")
    if args.fan_files:
        # Override given default content (if any)
        grammar, constraints = parse_contents_from_args(args)
    else:
        grammar = DEFAULT_FAN_CONTENT[0]
        constraints = DEFAULT_FAN_CONTENT[1]

    if grammar is None:
        raise FandangoError("Use '-f FILE.fan' to open a Fandango spec")

    # Avoid messing with default constraints
    constraints = constraints.copy()

    if DEFAULT_CONSTRAINTS:
        constraints += DEFAULT_CONSTRAINTS

    settings = make_fandango_settings(args, DEFAULT_SETTINGS)
    LOGGER.debug(f"Settings: {settings}")

    file_mode = get_file_mode(args, settings, grammar=grammar)
    LOGGER.info(f"File mode: {file_mode}")

    LOGGER.debug("Starting Fandango")
    fandango = Fandango._with_parsed(
        grammar,
        constraints,
        start_symbol=args.start_symbol,
        logging_level=LOGGER.getEffectiveLevel(),
    )
    LOGGER.debug("Evolving population")

    def solutions_callback(sol, i):
        return output_solution(sol, args, i, file_mode)

    max_generations = args.max_generations
    desired_solutions = args.desired_solutions
    infinite = args.infinite

    population = fandango.fuzz(
        solution_callback=solutions_callback,
        max_generations=max_generations,
        desired_solutions=desired_solutions,
        infinite=infinite,
        mode=FuzzingMode.COMPLETE,
        **settings,
    )

    if args.validate:
        LOGGER.debug("Validating population")

        # Ensure that every generated file can be parsed
        # and returns the same string as the original
        try:
            temp_dir = tempfile.TemporaryDirectory(delete=False)  # type: ignore [call-overload] # delete is only available on some OSs
        except TypeError:
            # Python 3.11 does not know the `delete` argument
            temp_dir = tempfile.TemporaryDirectory()
        args.directory = temp_dir.name
        args.format = "string"
        output_population(population, args, file_mode=file_mode, output_on_stdout=False)
        generated_files = glob.glob(args.directory + "/*")
        generated_files.sort()
        assert len(generated_files) == len(population)

        errors = 0
        for i in range(len(generated_files)):
            generated_file = generated_files[i]
            individual = population[i]

            try:
                with open_file(generated_file, file_mode, mode="r") as fd:
                    tree = parse_file(fd, args, grammar, constraints, settings)
                    validate(individual, tree, filename=fd.name)

            except Exception as e:
                print_exception(e)
                errors += 1

        if errors:
            raise FandangoError(f"{errors} error(s) during validation")

        # If everything went well, clean up;
        # otherwise preserve file for debugging
        shutil.rmtree(temp_dir.name)


def parse_command(args: argparse.Namespace) -> None:
    """Parse given files"""
    if args.fan_files:
        # Override given default content (if any)
        grammar, constraints = parse_contents_from_args(args)
    else:
        grammar = DEFAULT_FAN_CONTENT[0]
        constraints = DEFAULT_FAN_CONTENT[1]

    if grammar is None:
        raise FandangoError("Use '-f FILE.fan' to open a Fandango spec")

    # Avoid messing with default constraints
    constraints = constraints.copy()

    if DEFAULT_CONSTRAINTS:
        constraints += DEFAULT_CONSTRAINTS

    settings = make_fandango_settings(args, DEFAULT_SETTINGS)
    LOGGER.debug(f"Settings: {settings}")

    file_mode = get_file_mode(args, settings, grammar=grammar)
    LOGGER.info(f"File mode: {file_mode}")

    if not args.input_files:
        args.input_files = ["-"]

    population = []
    errors = 0

    for input_file in args.input_files:
        with open_file(input_file, file_mode, mode="r") as fd:
            try:
                tree = parse_file(fd, args, grammar, constraints, settings)
                population.append(tree)
            except Exception as e:
                print_exception(e)
                errors += 1
                tree = None

    if population and args.output:
        output_population(population, args, file_mode=file_mode, output_on_stdout=False)

    if errors:
        raise FandangoParseError(f"{errors} error(s) during parsing")


def talk_command(args: argparse.Namespace) -> None:
    """Interact with a program, client, or server"""
    # if not args.test_command and not args.client and not args.server:
    #     raise FandangoError(
    #         "Use '--client' or '--server' to create a client or server, "
    #         "or specify a command to interact with."
    #     )
    args.parties = []

    LOGGER.info("---------- Parsing FANDANGO content ----------")
    if args.fan_files:
        # Override given default content (if any)
        grammar, constraints = parse_contents_from_args(args)
    else:
        grammar = DEFAULT_FAN_CONTENT[0]
        constraints = DEFAULT_FAN_CONTENT[1]

    if grammar is None:
        raise FandangoError("Use '-f FILE.fan' to open a Fandango spec")

    if grammar.fuzzing_mode != FuzzingMode.IO:
        LOGGER.warning("Fandango spec does not specify interaction parties")

    # Avoid messing with default constraints
    constraints = constraints.copy()

    if DEFAULT_CONSTRAINTS:
        constraints += DEFAULT_CONSTRAINTS

    settings = make_fandango_settings(args, DEFAULT_SETTINGS)
    LOGGER.debug(f"Settings: {settings}")

    file_mode = get_file_mode(args, settings, grammar=grammar)
    LOGGER.info(f"File mode: {file_mode}")

    LOGGER.debug("Starting Fandango")

    fandango = Fandango._with_parsed(
        grammar=grammar,
        constraints=constraints,
        start_symbol=args.start_symbol,
        logging_level=LOGGER.getEffectiveLevel(),
    )
    LOGGER.debug("Evolving population")

    def solutions_callback(sol, i):
        return output_solution(sol, args, i, file_mode)

    max_generations = args.max_generations
    desired_solutions = args.desired_solutions
    infinite = args.infinite

    fandango.fuzz(
        solution_callback=solutions_callback,
        max_generations=max_generations,
        desired_solutions=desired_solutions,
        infinite=infinite,
        mode=FuzzingMode.IO,
        **settings,
    )


def convert_command(args: argparse.Namespace) -> None:
    """Convert a given language spec into Fandango .fan format"""

    output = args.output
    if output is None:
        output = sys.stdout

    for input_file in args.convert_files:
        from_format = args.from_format
        input_file_lower = input_file.lower()
        if from_format == "auto":
            if input_file_lower.endswith(".g4") or input_file_lower.endswith(".antlr"):
                from_format = "antlr"
            elif input_file_lower.endswith(".dtd"):
                from_format = "dtd"
            elif input_file_lower.endswith(".bt") or input_file_lower.endswith(".010"):
                from_format = "bt"
            elif input_file_lower.endswith(".fan"):
                from_format = "fan"
            else:
                raise FandangoError(
                    f"{input_file!r}: unknown file extension; use --from=FORMAT to specify the format"
                )

        temp_file = None
        if input_file == "-":
            # Read from stdin
            with open_file(input_file, "text", mode="r") as fd:
                contents = fd.read()
            temp_file = tempfile.NamedTemporaryFile(
                mode="w", delete=False, suffix=".tmp"
            )
            temp_file.write(contents)
            temp_file.flush()
            input_file = temp_file.name

        converter: FandangoConverter
        match from_format:
            case "antlr" | "g4":
                converter = ANTLRFandangoConverter(input_file)
                spec = converter.to_fan()
            case "dtd":
                converter = DTDFandangoConverter(input_file)
                spec = converter.to_fan()
            case "bt" | "010":
                if args.endianness == "little":
                    endianness = Endianness.LittleEndian
                else:
                    endianness = Endianness.BigEndian
                if args.bitfield_order == "left-to-right":
                    bitfield_order = BitfieldOrder.LeftToRight
                else:
                    bitfield_order = BitfieldOrder.RightToLeft

                converter = BTFandangoConverter(input_file)
                spec = converter.to_fan(
                    endianness=endianness, bitfield_order=bitfield_order
                )
            case "fan":
                converter = FandangoFandangoConverter(input_file, parties=args.parties)
                spec = converter.to_fan()

        print(spec, file=output, end="")
        if temp_file:
            # Remove temporary file
            temp_file.close()
            os.unlink(temp_file.name)

    if output != sys.stdout:
        output.close()


def clear_command(args: argparse.Namespace) -> None:
    CACHE_DIR = cache_dir()
    if args.dry_run:
        print(f"Would clear {CACHE_DIR}", file=sys.stderr)
    elif os.path.exists(CACHE_DIR):
        print(f"Clearing {CACHE_DIR}...", file=sys.stderr, end="")
        clear_cache()
        print("done", file=sys.stderr)


def nop_command(args: argparse.Namespace) -> None:
    # Dummy command such that we can list ! and / as commands. Never executed.
    pass


def copyright_command(args: argparse.Namespace) -> None:
    print("Copyright (c) 2024-2025 CISPA Helmholtz Center for Information Security.")
    print("All rights reserved.")


def version_command(args: argparse.Namespace) -> None:
    if sys.stdout.isatty():
        version_line = f"💃 {styles.color.ansi256(styles.rgbToAnsi256(128, 0, 0))}Fandango{styles.color.close} {fandango.version()}"
    else:
        version_line = f"Fandango {fandango.version()}"
    print(version_line)


COMMANDS: dict[str, Callable[[argparse.Namespace], None]] = {
    "set": set_command,
    "reset": reset_command,
    "fuzz": fuzz_command,
    "parse": parse_command,
    "talk": talk_command,
    "convert": convert_command,
    "clear-cache": clear_command,
    "cd": cd_command,
    "help": help_command,
    "copyright": copyright_command,
    "version": version_command,
    "exit": exit_command,
    "!": nop_command,
    "/": nop_command,
}


def run(command: Callable[[argparse.Namespace], None], args: argparse.Namespace) -> int:
    try:
        command(args)
    except Exception as e:
        print_exception(e)
        return 1

    return 0
